import numpy as np
import pandas as pd

__all__ = ['woe', 'iv']

def woe(y_true, feature, pos_label=1):
    """Calculate series woe value
    
    Parameters
    ----------
    y_true   : pd.Series, shape (n_samples,)
        The target variable for supervised learning problems.
    feature  : pd.Series, shape (n_samples,)
        x variable, model feature.
    pos_label : int, default=1
        positive label value.
    
    Returns
    -------
    t : series woe value
    """
    t = pd.DataFrame({'label':y_true, 'feature':feature})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    corr = t.label.sum()/(t.label.count()-t.label.sum())
    t = t.groupby(['feature']).label.apply(lambda x:np.log(x.sum()/(x.count()-x.sum())/corr))
    return t

def iv(y_true, feature, pos_label=1):
    """Calculate series iv value
    
    Parameters
    ----------
    y_true   : pd.Series, shape (n_samples,)
        The target variable for supervised learning problems.
    feature  : pd.Series, shape (n_samples,)
        x variable, model feature.
    pos_label : int, default=1
        positive label value.
    
    Returns
    -------
    t : series iv value
    """
    t = pd.DataFrame({'label':y_true, 'feature':feature})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    def _iv(group, df):
        a = group.sum()/df.label.sum()
        b = (group.count()-group.sum())/(t.label.count()-t.label.sum())
        return (a-b)/np.log(a/b)
    corr = t.label.sum()/(t.label.count()-t.label.sum())
    t = t.groupby(['feature']).label.apply(lambda x:_iv(x, t)).sum()
    return t



import numpy as np
import pandas as pd
import pyecharts as pe
from linora.metrics import confusion_matrix

__all__ = ['ks_curve', 'roc_curve', 'pr_curve', 'lift_curve', 'gain_curve', 'gini_curve',
           'confusion_matrix_map']

def ks_curve(y_true, y_pred, pos_label=1, jupyter=True, path='Kolmogorov-Smirnov Curve.html'):
    """plot Kolmogorov-Smirnov Curve.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        pos_label: positive label.
        jupyter: if plot in jupyter, default True.
        path: if jupyter is False, result save a html file.
    Returns:
        A pyecharts polt object.
    """
    t = pd.DataFrame({'prob':y_pred, 'label':y_true})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label'], ascending=False).reset_index(drop=True)
    t['tp'] = t.label.cumsum()
    t['fp'] = t.index+1-t.tp
    t['tpr'] = t.tp/t.label.sum()
    t['fpr'] = t.fp/(t.label.count()-t.label.sum())
    t['ks'] = (t.tpr-t.fpr).abs().round(4)
    t.index = np.round(((t.index+1)/len(t)), 2)
    line = (pe.charts.Line()
        .add_xaxis([0]+t.index.tolist())
        .add_yaxis("TPR", [0]+t.tpr.round(4).tolist(), is_smooth=True)
        .add_yaxis("FPR", [0]+t.fpr.round(4).tolist(), is_smooth=True)
        .add_yaxis("KS", [0]+t.ks.tolist(), is_smooth=True, markpoint_opts=pe.options.MarkPointOpts(data=[pe.options.MarkPointItem(type_='max', name='最大值')]))
        .set_series_opts(label_opts=pe.options.LabelOpts(is_show=False))
        .set_global_opts(title_opts=pe.options.TitleOpts(title="Kolmogorov-Smirnov Curve"),
                         xaxis_opts=pe.options.AxisOpts(is_scale=True, type_='value')))
    return line.render_notebook() if jupyter else line.render(path)

def roc_curve(y_true, y_pred, pos_label=1, jupyter=True, path='Receiver Operating Characteristic Curve.html'):
    """plot Receiver Operating Characteristic Curve.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        pos_label: positive label.
        jupyter: if plot in jupyter, default True.
        path: if jupyter is False, result save a html file.
    Returns:
        A pyecharts polt object.
    """
    t = pd.DataFrame({'prob':y_pred, 'label':y_true})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label'], ascending=False).reset_index(drop=True)
    t['tp'] = t.label.cumsum()
    t['fp'] = t.index+1-t.tp
    t['tpr'] = t.tp/t.label.sum()
    t['fpr'] = t.fp/(t.label.count()-t.label.sum())
    line = (pe.charts.Line()
        .add_xaxis([0]+t.fpr.round(4).tolist())
        .add_yaxis("ROC", [0]+t.tpr.round(4).tolist(), is_smooth=True)
        .add_yaxis("Random", [0]+t.fpr.round(4).tolist())
        .set_series_opts(label_opts=pe.options.LabelOpts(is_show=False),
                         areastyle_opts=pe.options.AreaStyleOpts(0.4))
        .set_global_opts(title_opts=pe.options.TitleOpts(title="ROC Curve"),
                         xaxis_opts=pe.options.AxisOpts('FPR', is_scale=True, type_='value'),
                         yaxis_opts=pe.options.AxisOpts('TPR')))
    return line.render_notebook() if jupyter else line.render(path)
    
def pr_curve(y_true, y_pred, pos_label=1, jupyter=True, path='Precision Recall Curve.html'):
    """plot Precision Recall Curve.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        pos_label: positive label.
        jupyter: if plot in jupyter, default True.
        path: if jupyter is False, result save a html file.
    Returns:
        A pyecharts polt object.
    """
    t = pd.DataFrame({'prob':y_pred, 'label':y_true})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label'], ascending=False).reset_index(drop=True)
    t['tp'] = t.label.cumsum()
    t['fp'] = t.index+1-t.tp
    t['recall'] = t.tp/t.label.sum()
    t['precision'] = t.tp/(t.tp+t.fp)
    line = (pe.charts.Line()
        .add_xaxis([0]+t.recall.round(4).tolist())
        .add_yaxis("PR", [1]+t.precision.round(4).tolist(), is_smooth=True)
        .set_series_opts(label_opts=pe.options.LabelOpts(is_show=False),
                         areastyle_opts=pe.options.AreaStyleOpts(0.4))
        .set_global_opts(title_opts=pe.options.TitleOpts(title="PR Curve"),
                         xaxis_opts=pe.options.AxisOpts('recall', is_scale=True, type_='value'),
                         yaxis_opts=pe.options.AxisOpts('precision')))
    return line.render_notebook() if jupyter else line.render(path)

def lift_curve(y_true, y_pred, pos_label=1, jupyter=True, path='Lift Curve.html'):
    """plot Lift Curve.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        pos_label: positive label.
        jupyter: if plot in jupyter, default True.
        path: if jupyter is False, result save a html file.
    Returns:
        A pyecharts polt object.
    """
    t = pd.DataFrame({'prob':y_pred, 'label':y_true})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label'], ascending=False).reset_index(drop=True)
    t['tp'] = t.label.cumsum()
    t['fp'] = t.index+1-t.tp
    t['precision'] = t.tp/(t.tp+t.fp)
    t['lift'] = t.precision/t.label.sum()*t.label.count()
    t.index = np.round(((t.index+1)/len(t)), 2)
    line = (pe.charts.Line()
        .add_xaxis(t.index.tolist())
        .add_yaxis("Lift", t.lift.round(3).tolist(), is_smooth=True)
        .add_yaxis("Random", [1]*len(t), is_smooth=True)
        .set_series_opts(label_opts=pe.options.LabelOpts(is_show=False))
        .set_global_opts(title_opts=pe.options.TitleOpts(title="Lift Curve"),
                         xaxis_opts=pe.options.AxisOpts(is_scale=True, type_='value')))
    return line.render_notebook() if jupyter else line.render(path)

def gain_curve(y_true, y_pred, pos_label=1, jupyter=True, path='Gain Curve.html'):
    """plot Gain Curve.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        pos_label: positive label.
        jupyter: if plot in jupyter, default True.
        path: if jupyter is False, result save a html file.
    Returns:
        A pyecharts polt object.
    """
    t = pd.DataFrame({'prob':y_pred, 'label':y_true})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label'], ascending=False).reset_index(drop=True)
    t['tp'] = t.label.cumsum()
    t['fp'] = t.index+1-t.tp
    t['precision'] = t.tp/(t.tp+t.fp)
    t.index = np.round(((t.index+1)/len(t)), 2)
    line = (pe.charts.Line()
        .add_xaxis(t.index.tolist())
        .add_yaxis("Gain", t.precision.round(4).tolist(), is_smooth=True)
        .set_series_opts(label_opts=pe.options.LabelOpts(is_show=False))
        .set_global_opts(title_opts=pe.options.TitleOpts(title="Gain Curve"),
                         xaxis_opts=pe.options.AxisOpts(is_scale=True, type_='value')))
    return line.render_notebook() if jupyter else line.render(path)

def gini_curve(y_true, y_pred, pos_label=1, jupyter=True, path='Gini Curve.html'):
    """plot Gini Curve.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        pos_label: positive label.
        jupyter: if plot in jupyter, default True.
        path: if jupyter is False, result save a html file.
    Returns:
        A pyecharts polt object.
    """
    t = pd.DataFrame({'prob':y_pred, 'label':y_true})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label']).reset_index(drop=True)
    t['cum'] = t.label.cumsum()/t.label.sum()
    t.index = np.round(((t.index+1)/len(t)), 2)
    line = (pe.charts.Line()
        .add_xaxis([0]+t.index.tolist())
        .add_yaxis("Gini", [0]+t.cum.round(4).tolist(), is_smooth=True)
        .add_yaxis("Random", [0]+t.index.tolist())
        .set_series_opts(label_opts=pe.options.LabelOpts(is_show=False),
                         areastyle_opts=pe.options.AreaStyleOpts(0.4))
        .set_global_opts(title_opts=pe.options.TitleOpts(title="Gini Curve"),
                         xaxis_opts=pe.options.AxisOpts(is_scale=True, type_='value')))
    return line.render_notebook() if jupyter else line.render(path)

def confusion_matrix_map(y_true, y_pred, jupyter=True, path="Confusion Matrix Map.html"):
    """plot Confusion Matrix Map.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        jupyter: if plot in jupyter, default True.
        path: if jupyter is False, result save a html file.
    Returns:
        A pyecharts polt object.
    """
    t = confusion_matrix(y_true, y_pred)
    t = pd.DataFrame([[i, m,n ] for i,j in t.to_dict().items() for m, n in j.items()],
                     columns=['actual', 'predict', 'over_values'])
    heatmap = (
        pe.charts.HeatMap()
        .add_xaxis(
            xaxis_data=t.drop_duplicates(['actual']).actual.values.tolist())
        .add_yaxis(
            "Confusion Matrix", 
            yaxis_data=t.drop_duplicates(['predict']).predict.values.tolist(), 
            value=t.values.tolist(),
            label_opts=pe.options.LabelOpts(is_show=True, color="#fff", position='inside', horizontal_align="50%")
        )
        .set_global_opts(
            title_opts=pe.options.TitleOpts(title="Confusion Matrix Map"),
            xaxis_opts=pe.options.AxisOpts(
                type_="category", 
                name='Actual', 
                splitarea_opts=pe.options.SplitAreaOpts(
                    is_show=True, areastyle_opts=pe.options.AreaStyleOpts(opacity=1))),
            yaxis_opts=pe.options.AxisOpts(
                type_="category", 
                name='Predict', 
                splitarea_opts=pe.options.SplitAreaOpts(
                    is_show=True, areastyle_opts=pe.options.AreaStyleOpts(opacity=1))),
            visualmap_opts=pe.options.VisualMapOpts(min_=0, max_=int(t.over_values.max()))))
    return  heatmap.render_notebook() if jupyter else heatmap.render(path)



import random

class Params:
    batch = 0
    batch_size = 1
    prefetch_size = 0
    repeat_size = 0
    shuffle_size = 0
    rank = dict()
    skip = 0
    take = -1
    shard_step = 1
    shard_index = 0

class DataSet():
    def __init__(self):
        self.params = Params()
    
    def batch(self, batch_size, drop_remainder=False):
        """Combines consecutive elements of this dataset into batches.
        Args:
            batch_size: representing the number of consecutive elements of this dataset to combine in a single batch.
            drop_remainder: representing whether the last batch should be dropped in the case it has fewer than batch_size elements; 
                            the default behavior is not to drop the smaller batch.
        Returns:
            A Dataset.
        """
        self.params.batch_size = batch_size
        self.params.drop_remainder = drop_remainder
        return self
        
    def prefetch(self, buffer_size):
        """Creates a Dataset that prefetches elements from this dataset.
        Args:
            buffer_size: representing the maximum number of elements that will be buffered when prefetching.
        Returns:
            A Dataset.
        """
        if 'prefetch' not in self.params.rank:
            self.params.rank['prefetch'] = len(self.params.rank)+1
        self.params.prefetch_size = buffer_size
        return self
        
    def repeat(self, count):
        """Repeats this dataset so each original value is seen count times.
        Args:
            count: representing the number of times the dataset should be repeated. 
        Returns:
            A Dataset.
        """
        if 'repeat' not in self.params.rank:
            self.params.rank['repeat'] = len(self.params.rank)+1
        self.params.repeat_size = count
        return self
        
    def shuffle(self, buffer_size, seed=None):
        """Randomly shuffles the elements of this dataset.
        Args:
            buffer_size: representing the number of elements from this dataset from which the new dataset will sample.
            seed: representing the random seed that will be used to create the distribution.
        Returns:
            A Dataset.
        """
        if 'shuffle' not in self.params.rank:
            self.params.rank['shuffle'] = len(self.params.rank)+1
        self.params.shuffle_size = buffer_size
        self.params.shuffle_seed = seed if seed is not None else random.randint(1, 99)
        return self
    
    def skip(self, count):
        """Creates a Dataset that skips count elements from this dataset.
        Args:
            count: representing the number of elements of this dataset that should be skipped to form the new dataset. 
                   If count is greater than the size of this dataset, 
                   the new dataset will contain no elements.
        Returns:
            A Dataset.
        """
        if 'skip' not in self.params.rank:
            self.params.rank['skip'] = len(self.params.rank)+1
        self.params.skip = count
        return self
        
    def take(self, count):
        """Creates a Dataset with at most count elements from this dataset.
        Args:
            count: representing the number of elements of this dataset that should be taken to form the new dataset. 
                   If count is -1, or if count is greater than the size of this dataset, 
                   the new dataset will contain all elements of this dataset.
        Returns:
            A Dataset.
        """
        if 'take' not in self.params.rank:
            self.params.rank['take'] = len(self.params.rank)+1
        self.params.take = count
        return self
    
    def shard(self, num_shards, index):
        """Creates a Dataset that includes only 1/num_shards of this dataset.
        Args:
            num_shards: representing the number of shards operating in parallel.
            index: representing the worker index.
        Returns:
            A Dataset.
        """
        if 'shard' not in self.params.rank:
            self.params.rank['shard'] = len(self.params.rank)+1
        self.params.shard_step = num_shards
        self.params.shard_index = index
        return self
    
    def map(self, map_func):
        """Maps map_func across the elements of this dataset.
        Args:
            map_func: A function mapping a dataset element to another dataset element.
        Returns:
            A Dataset.
        """
        if 'map' not in self.params.rank:
            self.params.rank['map'] = len(self.params.rank)+1
        self.params.map_func = map_func
        self.params.num_parallel_calls = num_parallel_calls
        self.params.deterministic = deterministic
        return self



from linora import chart
from linora import metrics
from linora import parallel
from linora import param_search
from linora import feature_column
from linora import feature_selection
from linora import sample
from linora import sample_splits
from linora import text
from linora import image
# from linora import database
from linora._logger import *
from linora._config import *

__version__ = '1.0.0rc3'
__author__ = 'JinQing Lee'

def _hello():
    print("""
------------------------------------------------------------------------------------
      Linora
--------------------
      Version      : --  {}  --
      Author       : --  {}  --
      License      : Apache-2.0
      Homepage     : https://github.com/Hourout/linora
      Description  : Simple and efficient tools for data mining and data analysis.
------------------------------------------------------------------------------------""".format(__version__, __author__))



from itertools import chain
import numpy as np

__all__ = ['select_best_length', 'word_to_index', 'word_index_sequence', 'pad_sequences',
           'index_vector_matrix']

def select_best_length(sequence, sample_rate=0.8):
    """Select best length for sequence with keep rate.
    
    Args:
        sequence: pd.Series or np.array or List of lists, sample sequence.
        sample_rate: float, keep sample rate.
    Returns:
        int length, sample max keep length.
    """
    t = sorted(map(lambda x:len(x), sequence))
    return t[int(np.ceil(len(sequence)*sample_rate))-1]

def word_to_index(sequence):
    """Sequence word transfer to index.
    
    Args:
        sequence: pd.Series or np.array or List of lists, sample sequence.
    Returns:
        dict, {word: index}.
    """
    return {v: k + 1 for k, v in enumerate(set(chain.from_iterable(sequence)))}

def word_index_sequence(sequence, word_index_dict, pad_value=0):
    """Sequence word transfer to sequence index.
    
    Args:
        sequence: pd.Series or np.array or List of lists, sample sequence.
        word_index_dict: dict, {word: index}.
        pad_value: fillna value, if word not in word_index_dict, fillna it.
    Returns:
        List of lists, sequence word transfer to sequence index list.
    """
    t = []
    for i in sequence:
        index = []
        for j in i:
            try:
                index.append(word_index_dict[j])
            except:
                index.append(pad_value)
        t.extend([index])
    return t

def pad_sequences(sequence, maxlen=None, dtype='int32', padding='pre', truncating='pre', value=0):
    """Pads sequences to the same length.
    
    Args:
        sequences: pd.Series or np.array or List of lists, where each element is a sequence.
        maxlen: Int, maximum length of all sequences.
        dtype: Type of the output sequences.
               To pad sequences with variable length strings, you can use `object`.
        padding: String, 'pre' or 'post':
                 pad either before or after each sequence.
        truncating: String, 'pre' or 'post':
                    remove values from sequences larger than
                    `maxlen`, either at the beginning or at the end of the sequences.
        value: Float or String, padding value.
    Returns:
        List of lists with shape `(len(sequences), maxlen)`
    """
    t = []
    if padding=='post' and truncating=='post':
        for i in sequence:
            t.extend([i[:maxlen] if len(i)>maxlen else i+[value]*(maxlen-len(i))])
    elif padding=='post' and truncating=='pre':
        for i in sequence:
            t.extend([i[-maxlen:] if len(i)>maxlen else i+[value]*(maxlen-len(i))])
    elif padding=='pre' and truncating=='post':
        for i in sequence:
            t.extend([i[:maxlen] if len(i)>maxlen else [value]*(maxlen-len(i))+i])
    elif padding=='pre' and truncating=='pre':
        for i in sequence:
            t.extend([i[-maxlen:] if len(i)>maxlen else [value]*(maxlen-len(i))+i])
    else:
        raise ValueError('Padding type "%s" not understood or Truncating type "%s" not understood' % (padding, truncating))
    return t

def index_vector_matrix(word_index_dict, word_vector_dict, embed_dim=300, initialize='norm', dtype='float32'):
    """Make index vector matrix with shape `(len(word_index_dict), embed_dim)`.
    
    Args:
        word_index_dict: dict, {word: index}.
        word_vector_dict: dict, {word: vector}.
        embed_dim: embedding dim.
        initialize: initialize method, 'zero' or 'one' or 'norm'.
        dtype: default 'float32', data types.
    Returns:
        a index vector matrix with shape `(len(word_index_dict), embed_dim)`.
    """
    word_count = len(word_index_dict)+1
    init = {'zero':lambda x,y:np.zeros((x,y), dtype=dtype),
            'one':lambda x,y:np.ones((x,y), dtype=dtype),
            'norm':lambda x,y:np.random.normal(size=(x,y), dtype=dtype)}
    index_embed_matrix = init[initialize](word_count, embed_dim)
    for word, index in word_index_dict.items():
        l = len(word_vector_dict[word])
        if l>embed_dim:
            index_embed_matrix[index, :] = word_vector_dict[word][:embed_dim]
        else:
            index_embed_matrix[index, :l] = word_vector_dict[word]
    return index_embed_matrix



from linora.image._image_resize import *

class ImageResizeAug(object):
    def __init__(self, image=None):
        self.image = image
        
    def resize(self, size, method=ResizeMethod.BILINEAR):
        """Returns a resized copy of this image.
    
        Args:
        size: The requested size in pixels, as a 2-tuple: (width, height).
        method: An optional resampling filter. see la.image.ResizeMethod.

        Returns:
            a Image instance.
        """
        self.image = resize(self.image, size, method)
        return self
    
    



import numpy as np
from linora.image._image_io import *
from linora.image._image_util import *

class ImageIoAug(object):
    def __init__(self, image=None):
        self.image = image
    
    def read_image(self, filename):
        """
        Args:
            filename: str, image absolute path.
        Returns:
            a Image instance.
        """
        self.image = read_image(filename)
        return self
    
    def save_image(self, filename, file_format=None, **kwargs):
        """Saves an image stored as a Numpy array to a path or file object.
        Args
            filename: Path or file object.
            image: A PIL Image instance.
            file_format: Optional file format override. If omitted, the
                format to use is determined from the filename extension.
                If a file object was used instead of a filename, this
                parameter should always be used.
            **kwargs: Additional keyword arguments passed to `PIL.Image.save()`.
        """
        return save_image(filename, self.image, file_format=file_format, **kwargs)
    
    def color_convert(self, color_mode='RGB'):
        """Transform image color mode

        Args
            img: PIL Image instance.
            color_mode: Image color mode, more see la.image.ColorMode
        Returns
            PIL Image instance.
        Raises
            ValueError: color_mode error.
        """
        self.image = color_convert(self.image, color_mode=color_mode)
        return self
    
    def image_to_array(self, data_format='channels_last', dtype='float32'):
        """Converts a PIL Image instance to a Numpy array.
        
        Args
            img: PIL Image instance.
            data_format: Image data format,
                either "channels_first" or "channels_last".
            dtype: Dtype to use for the returned array.
        Returns
            A 3D Numpy array.
        Raises
            ValueError: if invalid `img` or `data_format` is passed.
        """
        self.image = image_to_array(self.image, data_format, dtype)
        return self
        
    def array_to_image(self, data_format='channels_last'):
        """Converts a 3D Numpy array to a PIL Image instance.
        Args
            x: Input Numpy array.
            data_format: Image data format, either "channels_first" or "channels_last".
                Default: "channels_last".
        Returns
            A PIL Image instance.
        Raises
            ValueError: if invalid `x` or `data_format` is passed.
        """
        self.image = array_to_image(self.image, data_format)
        return self



import numpy as np

__all__ = ['noise_gaussian', 'noise_poisson', 'noise_mask',
           'noise_saltpepper', 'noise_rainbow']

def noise_gaussian(image, scale=1, mean=0.0, std=1.0):
    """Gaussian noise apply to image.
    
    new pixel = image + gaussian_noise * scale
    Args:
        image: Either a 3-D float Tensor of shape [height, width, depth].
        scale: if int or float, value multiply with poisson_noise.
               if tuple or list, randomly picked in the interval
               `[scale[0], scale[1])`, value multiply with poisson_noise.
        mean: if int or float, value is gaussian distribution mean.
              if tuple or list, randomly picked in the interval
              `[mean[0], mean[1])`, value is gaussian distribution mean.
        std: if int or float, value is gaussian distribution std.
             if tuple or list, randomly picked in the interval
             `[std[0], std[1])`, value is gaussian distribution std.
    Returns:
        3-D float Tensor, as per the input.
    Raises:
        scale or lam type error.
    """
    if isinstance(scale, (tuple, list)):
        scale = np.random.uniform(scale[0], scale[1])
    if isinstance(std, (tuple, list)):
        std = np.random.uniform(std[0], std[1])
    if isinstance(mean, (tuple, list)):
        mean = np.random.uniform(mean[0], mean[1])
    return np.random.normal(mean, std, size=image.size)*scale+image

def noise_poisson(image, scale=1, lam=1.0):
    """Poisson noise apply to image.
    
    new pixel = image + poisson_noise * scale
    Args:
        image: Either a 3-D float Tensor of shape [height, width, depth].
        scale: if int or float, value multiply with poisson_noise.
               if tuple or list, randomly picked in the interval
               `[scale[0], scale[1])`, value multiply with poisson_noise.
        lam: if int or float, value is poisson distribution lambda.
             if tuple or list, randomly picked in the interval
             `[lam[0], lam[1])`, value is poisson distribution lambda.
    Returns:
        3-D float Tensor, as per the input.
    Raises:
        scale or lam type error.
    """
    if isinstance(scale, (tuple, list)):
        scale = np.random.uniform(scale[0], scale[1])
    if isinstance(lam, (tuple, list)):
        lam = np.random.uniform(std[0], std[1])
    return np.random.poisson(lam, size=image.size)*scale+image

def noise_mask(image, noise_prob=0.2):
    """Mask noise apply to image.
    
    With probability `drop_prob`, outputs the input element scaled up by
    `1`, otherwise outputs `0`. 
    
    Tips:
        1 mean pixel have no change.
        a suitable interval is (0., 0.1].
    Args:
        image: Either a 3-D float Tensor of shape [height, width, depth].
        noise_prob: should be in the interval (0, 1.].
                   if float, the probability that each element is drop.
                   if tuple or list, randomly picked in the interval
                   `[keep_prob[0], keep_prob[1])`, the probability that each element is drop.
    Returns:
        3-D float Tensor, as per the input.
    Raises:
        ValueError: If `keep_prob` is not in `(0, 1.]`.
    """
    if isinstance(noise_prob, (tuple, list)):
        noise_prob = random.uniform(noise_prob[0], noise_prob[1])
    return np.random.choice([0, 1], size=image.size, p=[noise_prob,1-noise_prob])*image

def noise_saltpepper(image, noise_prob=0.2):
    """Salt-Pepper noise apply to image.
    
    The salt-pepper noise is based on the signal-to-noise ratio of the image,
    randomly generating the pixel positions in some images all channel,
    and randomly assigning these pixels to 0 or 255.
    
    Args:
        image: Either a 3-D float Tensor of shape [height, width, depth].
        noise_prob: should be in the interval (0, 1].
                   if int or float, the probability that each element is kept.
                   if tuple or list, randomly picked in the interval
                   `[keep_prob[0], keep_prob[1])`, the probability that each element is kept.
    Returns:
        3-D float Tensor, as per the input.
    Raises:
        ValueError: If `keep_prob` is not in `(0, 1]`.
    """
    if isinstance(noise_prob, (tuple, list)):
        noise_prob = random.uniform(noise_prob[0], noise_prob[1])
    shape = [i for i in image.shape if i!=min(image.shape)]
    noise = np.random.choice([0, 1, 2], size=image.size, p=[noise_prob/2, 1-noise_prob, noise_prob/2])
    image[noise>1] = 255
    image[noise<1] = 0
    return image

def noise_rainbow(image, noise_prob=0.2):
    """Rainbowr noise apply to image.
    
    The rainbow noise is based on the signal-to-noise ratio of the image,
    randomly generating the pixel positions in some images,
    and randomly assigning these pixels to 0 or 255.
    
    Args:
        image: Either a 3-D float Tensor of shape [height, width, depth].
        noise_prob: should be in the interval (0, 1].
                   if int or float, the probability that each element is kept.
                   if tuple or list, randomly picked in the interval
                   `[keep_prob[0], keep_prob[1])`, the probability that each element is kept.
    Returns:
        3-D float Tensor, as per the input.
    Raises:
        ValueError: If `keep_prob` is not in `(0, 1]`.
    """
    if isinstance(noise_prob, (tuple, list)):
        noise_prob = random.uniform(noise_prob[0], noise_prob[1])
    noise = np.random.choice([0, 1, 2], size=image.shape, p=[noise_prob/2, 1-noise_prob, noise_prob/2])
    image[noise>1] = 255
    image[noise<1] = 0
    return image



# from linora.param_search import GEClassifier
# from linora.param_search import GERegressor
# from linora.param_search import LGBClassifier
from linora.param_search import XGBClassifier
from linora.param_search import XGBRanker
from linora.param_search import XGBRegressor

from linora.param_search._HyperParameters import *

from linora.param_search._config import __xgboost_version__



from linora.image._image_filter import *

class ImageFilterAug(object):
    def __init__(self, image=None):
        self.image = image
    
    def filter_BoxBlur(self, radius=2):
        """
        Blurs the image by setting each pixel to the average value of the pixels 
        in a square box extending radius pixels in each direction. 
        Supports float radius of arbitrary size. Uses an optimized implementation 
        which runs in linear time relative to the size of the image for any radius value.

        Args:
        radius: Size of the box in one direction. Radius 0 does not blur, returns an identical image. 
                Radius 1 takes 1 pixel in each direction, i.e. 9 pixels in total.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_BoxBlur(self.image, radius)
        return self

    def filter_GaussianBlur(self, radius=2):
        """Gaussian blur filter.

        Args:
        radius: Blur radius.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_GaussianBlur(self.image, radius)
        return self

    def filter_UnsharpMask(self, radius=2, percent=150, threshold=3):
        """Unsharp mask filter.
        See Wikipedia’s entry on digital unsharp masking for an explanation of the parameters.

        Args:
        radius: Blur Radius
        percent: Unsharp strength, in percent
        threshold: Threshold controls the minimum brightness change that will be sharpened

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_UnsharpMask(self.image, radius, percent, threshold)
        return self

    def filter_Rank(self, size, rank):
        """Create a rank filter. 
        The rank filter sorts all pixels in a window of the given size, and returns the rank’th value.

        Args:
        size: The kernel size, in pixels.
        rank: What pixel value to pick. Use 0 for a min filter, 
              size * size / 2 for a median filter, size * size - 1 for a max filter, etc.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_Rank(self.image, size, rank)
        return self

    def filter_Median(self, size=3):
        """Create a median filter. Picks the median pixel value in a window with the given size.

        Args:
        size: The kernel size, in pixels.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_Median(self.image, size)
        return self

    def filter_Min(self, size=3):
        """Create a min filter. Picks the lowest pixel value in a window with the given size.

        Args:
        size: The kernel size, in pixels.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_Min(self.image, size)
        return self

    def filter_Max(self, size=3):
        """Create a max filter. Picks the largest pixel value in a window with the given size.

        Args:
        size: The kernel size, in pixels.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_Max(self.image, size)
        return self

    def filter_Mode(self, size=3):
        """Create a mode filter. Picks the most frequent pixel value in a box with the given size. 

        Pixel values that occur only once or twice are ignored; 
        if no pixel value occurs more than twice, 
        the original pixel value is preserved.

        Args:
        size: The kernel size, in pixels.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_Mode(self.image, size)
        return self

    def filter_BLUR(self):
        """Normal blur.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_BLUR(self.image)
        return self

    def filter_CONTOUR(self):
        """contour blur.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_CONTOUR(self.image)
        return self

    def filter_DETAIL(self):
        """detall blur.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_DETAIL(self.image)
        return self

    def filter_EDGE_ENHANCE(self):
        """Edge enhancement blur.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_EDGE_ENHANCE(self.image)
        return self

    def filter_EDGE_ENHANCE_MORE(self):
        """Edge enhancement threshold blur.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_EDGE_ENHANCE_MORE(self.image)
        return self

    def filter_EMBOSS(self):
        """emboss blur.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_EMBOSS(self.image)
        return self

    def filter_FIND_EDGES(self):
        """Find the edge blur.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_FIND_EDGES(self.image)
        return self

    def filter_SHARPEN(self):
        """Sharpen blur.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_SHARPEN(self.image)
        return self

    def filter_SMOOTH(self):
        """Smooth blur.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_SMOOTH(self.image)
        return self

    def filter_SMOOTH_MORE(self):
        """Smooth threshold blur.

        Returns:
                A Image instance. of the same type and shape as `image`. 
        """
        self.image = filter_SMOOTH_MORE(self.image)
        return self



import os
import pathlib
from itertools import chain

import pandas as pd

__all__ = ['ImageDataset', 'ImageClassificationFolderDataset']


class ImageDataset():
    def __init__(self, root, image_format=['png', 'jpg', 'jpeg'], label_func=None):
        """Construct an image dataset label index.
        
        Args:
            root: image dataset file root.
            image_format: list, default ['png', 'jpg', 'jpeg'].
            label_func: if label_func is None, self.data['label'] is not exist; function is apply to self.data['image'].
        Returns:
            class, self.data['image'] is image path, self.data['label'] is image label.
        """
        self.root = root
        self.image_format = image_format

        p = pathlib.Path(self.root)
        self.data = pd.DataFrame(chain.from_iterable((p.rglob(f'*.{i}') for i in self.image_format)), columns=['image'])
        if label_func is not None:
            self.data['label'] = self.data.image.map(lambda x:label_func(x.name))
        self.data['image'] = self.data.astype(str).image.map(lambda x:eval(repr(x).replace("\\", '/').replace("//", '/')))

class ImageClassificationFolderDataset():
    def __init__(self, root, image_format=['png', 'jpg', 'jpeg'], label_encoder=False):
        """Construct an image dataset label index.
        
        Args:
            root: image dataset file root.
            image_format: list, default ['png', 'jpg', 'jpeg'].
            label_encoder: whether encode labels with value between 0 and n_classes-1.
        Returns:
            class, self.data['image'] is image path, self.data['label'] is image label.
        """
        self.root = root
        self.image_format = image_format    

        file = os.listdir(self.root)
        file = [i for i in file if os.path.isdir(self.root+'/'+i) and i[0]!='.']
        data = pd.DataFrame()
        for i in file:
            data = pd.concat([data, pd.DataFrame({'image':os.listdir(self.root+'/'+i), 'label':i})])
        data = data.reset_index(drop=True)
        data['image'] = self.root+'/'+data.label+'/'+data.image
        data = data[data.image.map(lambda x: True if '.' in x.split('/')[-1] else False)]
        data = data[data.image.map(lambda x: True if x.split('/')[-1][0]!='.' else False)]
        data = data[data.image.map(lambda x: True if len(x.split('/')[-1].split('.'))==2 else False)]
        data = data[data.image.map(lambda x: True if str.lower(x.split('/')[-1].split('.')[1]) in self.image_format else False)]
        self.dataset = data.reset_index(drop=True)
        self.name_label_dict = {j: i for i, j in enumerate(data.label.unique())}
        if label_encoder:
            self.dataset['label'] = self.dataset.label.replace(self.name_label_dict)



import pyecharts as pe

__all__ = ['feature_corr_map']

def feature_corr_map(df, jupyter=True, path="Feature_Correlation_Map.html"):
    t = [[nor, row, j] for nor, i in enumerate(df.corr().round(3).values.tolist()) for row, j in enumerate(i)]
    heatmap = (
            pe.charts.HeatMap()
            .add_xaxis(
                xaxis_data=df.columns.tolist())
            .add_yaxis(
                "Feature Correlation", 
                yaxis_data=df.columns.tolist(), 
                value=t,
                label_opts=pe.options.LabelOpts(is_show=True, color="#fff", position='inside', horizontal_align="50%")
            )
            .set_global_opts(
                title_opts=pe.options.TitleOpts(title="Feature Correlation Map"),
                xaxis_opts=pe.options.AxisOpts(
                    type_="category", 
                    name='Actual', 
                    splitarea_opts=pe.options.SplitAreaOpts(
                        is_show=True, areastyle_opts=pe.options.AreaStyleOpts(opacity=1))),
                yaxis_opts=pe.options.AxisOpts(
                    type_="category", 
                    name='Predict', 
                    splitarea_opts=pe.options.SplitAreaOpts(
                        is_show=True, areastyle_opts=pe.options.AreaStyleOpts(opacity=1))),
                visualmap_opts=pe.options.VisualMapOpts(min_=0, max_=int(a.max().max()))))
    return  heatmap.render_notebook() if jupyter else heatmap.render(path)



import numpy as np

__all__ = ['normalize_meanminmax', 'normalize_minmax', 'normalize_maxabs', 'normalize_max',
           'normalize_l1', 'normalize_l2', 'normalize_norm', 'normalize_robust']

def normalize_meanminmax(feature, feature_scale=None):
    """normalize feature with meanminmax method.
    
    Args:
        feature: pd.Series, sample feature value.
        feature_scale: list or tuple, [feature.mean(), feature.min(), feature.max()].
    Returns:
        normalize feature and feature_scale.
    """
    scale = feature_scale if feature_scale is not None else (feature.mean(), feature.min(), feature.max())
    t = (feature-scale[0])/(scale[2]-scale[1])
    return t, scale

def normalize_minmax(feature, feature_range=(0, 1), feature_scale=None):
    """normalize feature with minmax method.
    
    Args:
        feature: pd.Series, sample feature value.
        feature_range: list or tuple, range of values after feature transformation.
        feature_scale: list or tuple, [feature.min(), feature.max()].
    Returns:
        normalize feature and feature_scale.
    """
    scale = feature_scale if feature_scale is not None else (feature.min(), feature.max())
    t = (feature-scale[0])/(scale[1]-scale[0])*(feature_range[1]-feature_range[0])+feature_range[0]
    return t, scale

def normalize_maxabs(feature, feature_scale=None):
    """normalize feature with maxabs method.
    
    Args:
        feature: pd.Series, sample feature value.
        feature_scale: int or float, feature.abs().max().
    Returns:
        normalize feature and feature_scale.
    """
    scale = abs(feature_scale) if feature_scale is not None else feature.abs().max()
    t = feature/scale
    return t, scale

def normalize_max(feature, feature_scale=None):
    """normalize feature with max method.
    
    Args:
        feature: pd.Series, sample feature value.
        feature_scale: int or float, feature.abs().max().
    Returns:
        normalize feature and feature_scale.
    """
    scale = feature_scale if feature_scale is not None else feature.max()
    t = feature/scale
    return t, scale

def normalize_l1(feature, feature_scale=None):
    """normalize feature with l1 method.
    
    Args:
        feature: pd.Series, sample feature value.
        feature_scale: int or float, feature.abs().sum().
    Returns:
        normalize feature and feature_scale.
    """
    scale = feature_scale if feature_scale is not None else feature.abs().sum()
    t = feature/scale
    return t, scale

def normalize_l2(feature, feature_scale=None):
    """normalize feature with l2 method.
    
    Args:
        feature: pd.Series, sample feature value.
        feature_scale: int or float, np.sqrt(np.sum(np.square(feature))).
    Returns:
        normalize feature and feature_scale.
    """
    scale = feature_scale if feature_scale is not None else np.sqrt(np.sum(np.square(feature)))
    t = feature/scale
    return t, scale

def normalize_norm(feature, feature_scale=None):
    """normalize feature with norm method.
    
    Args:
        feature: pd.Series, sample feature value.
        feature_scale: list or tuple, [feature.mean(), feature.std()].
    Returns:
        normalize feature and feature_scale.
    """
    scale = feature_scale if feature_scale is not None else (feature.mean(), feature.std())
    t = (feature-scale[0])/scale[1]
    return t, scale

def normalize_robust(feature, feature_scale=(None, 0.5)):
    """normalize feature with robust method.
    
    Args:
        feature: pd.Series, sample feature value.
        feature_scale: list or tuple, [feature.median(), feature.quantile(0.75)-feature.quantile(0.25)];
                       if feature_scale[0] is not None, 
                       (feature_scale[0], feature.quantile(0.5+feature_scale[1]/2)-feature.quantile(0.5-feature_scale[1]/2)).
    Returns:
        normalize feature and feature_scale.
    """
    if feature_scale[0] is not None:
        scale = (feature_scale[0], feature.quantile(0.5+feature_scale[1]/2)-feature.quantile(0.5-feature_scale[1]/2))
    else:
        scale = (feature.median(), feature.quantile(0.75)-feature.quantile(0.25))
    t = (feature-scale[0])/scale[1]
    return t, scale



import os
import imp
import json


__all__ = ['Config']

def Config(file_py=None, file_dict=None, file_json=None, **kwargs):
    class config:
        if file_dict is not None:
            for i, j in file_dict.items():
                locals()[i] = j
        if file_json is not None:
            with open(file_json) as _f:
                data = json.load(_f)
            for i, j in data.items():
                locals()[i] = j
            try:
                del data, _f
            except:
                pass
        if file_py is not None:
            filters = ['__builtins__', '__cached__', '__doc__', '__file__', 
                       '__loader__', '__name__', '__package__', '__spec__']
            temp = imp.load_source(os.path.split(file_py)[-1].split('.')[0], file_py)
            for i in dir(temp):
                if i not in filters:
                    locals()[i] = getattr(temp, i)
            try:
                del temp, filters
            except:
                pass
        for i, j in kwargs.items():
            locals()[i] = j
        try:
            del i,j
        except:
            pass
    return config()



import numpy as np
from PIL import Image

__all__ = ['crop_central', 'crop_point']

def crop_central(image, central_rate):
    """Crop the central region of the image.
    
    Remove the outer parts of an image but retain the central region of the image
    along each dimension. If we specify central_fraction = 0.5, this function
    returns the region marked with "X" in the below diagram.
       --------
      |        |
      |  XXXX  |
      |  XXXX  |
      |        |   where "X" is the central 50% of the image.
       --------
  
    This function works on either a single image (`image` is a 3-D Tensor), or a
    batch of images (`image` is a 4-D Tensor).
    
    Args:
        image: a Image instance.
        central_rate: if int float, should be in the interval (0, 1], fraction of size to crop.
                      if tuple list, randomly picked in the interval
                      `[central_rate[0], central_rate[1])`, value is fraction of size to crop.
    Returns:
        a Image instance.
    Raises:
        ValueError: if central_crop_fraction is not within (0, 1].
    """
    if isinstance(central_rate, (int, float)):
        assert 0<central_rate<=1, 'if central_rate type one of int or float, must be in the interval (0, 1].'
    elif isinstance(central_rate, (tuple, list)):
        assert 0<central_rate[0]<central_rate[1]<=1, 'central_rate should be 1 >= central_rate[1] > central_rate[0] > 0.'
        central_rate = np.random.uniform(central_rate[0], central_rate[1])
    else:
        raise ValueError('central_rate should be one of int, float, tuple, list.')
    left = int(im.size[0]*(0.5-central_rate/2))
    upper = int(im.size[1]*(0.5-central_rate/2))
    right = int(im.size[0]*(0.5+central_rate/2))
    lower = int(im.size[1]*(0.5+central_rate/2))
    return image.crop((left, upper, right, lower))

def crop_point(image, height_rate, width_rate):
    """Crop the any region of the image(s) and resize specify shape.
    
    Crop region area = height_rate * width_rate *image_height * image_width
    
    This function works on either a single image (`image` is a 3-D Tensor), or a
    batch of images (`image` is a 4-D Tensor).
    
    Args:
        image: a Image instance.
        height_rate: flaot, in the interval (0, 1].
        width_rate: flaot, in the interval (0, 1].
    Returns:
        a Image instance.
    Raises:
        ValueError: if central_crop_fraction is not within (0, 1].
    """
    assert isinstance(height_rate, (int, float)), 'height_rate should be one of int, float.'
    assert isinstance(width_rate, (int, float)), 'width_rate should be one of int, float.'
    assert 0<height_rate<=1 and 0<width_rate<=1, 'height_rate and width_rate should be in the interval (0, 1].'
    left = image.size[0]*np.random.uniform(0, 1-width_rate)
    upper = image.size[1]*np.random.uniform(0, 1-height_rate)
    right = left+image.size[0]*width_rate
    lower = upper+image.size[1]*height_rate
    return image.crop((left, upper, right, lower))



import os

import numpy as np
from PIL import Image

__all__ = ['list_images', 'ColorMode', 'color_convert', 'image_to_array', 'array_to_image']

def list_images(directory, file_format=('jpg', 'jpeg', 'bmp', 'png', 'ppm', 'tif', 'tiff')):
    """Lists all pictures in a directory, including all subdirectories.
    Args:
        directory: string, absolute path to the directory
        file_format: tuple of strings or single string, extensions of the pictures
    Returns:
        a list of paths
    """
    file_format = tuple('.%s' % e for e in ((file_format,) if isinstance(file_format, str) else file_format))
    return [os.path.join(root, f) for root, _, files in os.walk(directory) for f in files if f.lower().endswith(file_format)]

class color_mode:
    O     = {'mode':'1',      'description':'1-bit pixels, black and white, stored with one pixel per byte'}
    L     = {'mode':'L',      'description':'8-bit pixels, black and white'}
    LA    = {'mode':'LA',     'description':'L with alpha'}
    La    = {'mode':'La',     'description':'L with premultiplied alpha'}
    I     = {'mode':'I',      'description':'32-bit signed integer pixels'}
    F     = {'mode':'F',      'description':'32-bit floating point pixels'}
    P     = {'mode':'P',      'description':'8-bit pixels, mapped to any other mode using a color palette'}
    PA    = {'mode':'PA',     'description':'P with alpha'}
    RGB   = {'mode':'RGB',    'description':'3x8-bit pixels, true color'}
    RGBA  = {'mode':'RGBA',   'description':'4x8-bit pixels, true color with transparency mask'}
    RGBX  = {'mode':'RGBX',   'description':'true color with padding'}
    RGBa  = {'mode':'RGBa',   'description':'true color with premultiplied alpha'}
    CMYK  = {'mode':'CMYK',   'description':'4x8-bit pixels, color separation'}
    YCbCr = {'mode':'YCbCr',  'description':'3x8-bit pixels, color video format'}
    LAB   = {'mode':'LAB',    'description':'3x8-bit pixels, the L*a*b color space'}
    HSV   = {'mode':'HSV',    'description':'3x8-bit pixels, Hue, Saturation, Value color space'}
    I16   = {'mode':'I;16',   'description':'16-bit unsigned integer pixels'}
    I16L  = {'mode':'I;16L',  'description':'16-bit little endian unsigned integer pixels'}
    I16B  = {'mode':'I;16B',  'description':'16-bit big endian unsigned integer pixels'}
    I16N  = {'mode':'I;16N',  'description':'16-bit native endian unsigned integer pixels'}
    BGR15 = {'mode':'BGR;15', 'description':'15-bit reversed true colour'}
    BGR16 = {'mode':'BGR;16', 'description':'16-bit reversed true colour'}
    BGR24 = {'mode':'BGR;24', 'description':'24-bit reversed true colour'}
    BGR32 = {'mode':'BGR;32', 'description':'32-bit reversed true colour'}

ColorMode = color_mode()

def color_convert(image, color_mode=ColorMode.RGB):
    """Transform image color mode
    
    Args
        img: PIL Image instance.
        color_mode: Image color mode, more see la.image.ColorMode
    Returns
        PIL Image instance.
    Raises
        ValueError: color_mode error.
    """
    if color_mode == 'grayscale':
        if image.mode not in ('L', 'I;16', 'I'):
            image = image.convert('L')
    elif isinstance(color_mode, dict):
        image = image.convert(color_mode['mode'])
    elif isinstance(color_mode, str):
        image = image.convert(color_mode)
    else:
        raise ValueError('color_mode error.')
    return image

def image_to_array(image, data_format='channels_last', dtype='float32'):
    """Converts a PIL Image instance to a Numpy array.
    Args
        img: PIL Image instance.
        data_format: Image data format,
            either "channels_first" or "channels_last".
        dtype: Dtype to use for the returned array.
    Returns
        A 3D Numpy array.
    Raises
        ValueError: if invalid `img` or `data_format` is passed.
    """
    if data_format not in {'channels_first', 'channels_last'}:
        raise ValueError('Unknown data_format: %s' % data_format)
    x = np.asarray(image, dtype=dtype)
    if len(x.shape) == 3:
        if data_format == 'channels_first':
            x = x.transpose(2, 0, 1)
    elif len(x.shape) == 2:
        if data_format == 'channels_first':
            x = x.reshape((1, x.shape[0], x.shape[1]))
        else:
            x = x.reshape((x.shape[0], x.shape[1], 1))
    else:
        raise ValueError('Unsupported image shape: %s' % (x.shape,))
    return x

def array_to_image(x, data_format='channels_last'):
    """Converts a 3D Numpy array to a PIL Image instance.
    Args
        x: Input Numpy array.
        data_format: Image data format, either "channels_first" or "channels_last".
            Default: "channels_last".
    Returns
        A PIL Image instance.
    Raises
        ValueError: if invalid `x` or `data_format` is passed.
    """
    if x.ndim != 3:
        raise ValueError('Expected image array to have rank 3 (single image). '
                         'Got array with shape: %s' % (x.shape,))

    if data_format not in {'channels_first', 'channels_last'}:
        raise ValueError('Invalid data_format: %s' % data_format)

    if data_format == 'channels_first':
        x = x.transpose(1, 2, 0)

    if x.shape[2] == 4:
        return Image.fromarray(x.astype('uint8'), 'RGBA')
    elif x.shape[2] == 3:
        return Image.fromarray(x.astype('uint8'), 'RGB')
    elif x.shape[2] == 1:
        if np.max(x) > 255:
            return Image.fromarray(x[:, :, 0].astype('int32'), 'I')
        return Image.fromarray(x[:, :, 0].astype('uint8'), 'L')
    else:
        raise ValueError('Unsupported channel number: %s' % (x.shape[2],))



from linora.image._image_color import *

class ImageColorAug(object):
    def __init__(self, image=None):
        self.image = image
    
    def enhance_color(self, delta):
        """Adjust image color balance.
        This class can be used to adjust the colour balance of an image, 
        in a manner similar to the controls on a colour TV set. 
        An enhancement factor of 0.0 gives a black and white image. 
        A factor of 1.0 gives the original image.

        Args:
        image: a Image instance.
        delta: A floating point value controlling the enhancement. 
               delta 1.0 always returns a copy of the original image, 
               lower factors mean less color, 
               and higher values more. There are no restrictions on this value.
               if list, tuple, randomly picked in the interval
                   `[delta[0], delta[1])` , value is float multiplier for adjusting color.
        Returns:
            a Image instance.
        """
        self.image = enhance_color(self.image, delta)
        return self
    
    def enhance_contrast(self, delta):
        """Adjust contrast of RGB or grayscale images.
  
        Contrast is adjusted independently for each channel of each image.

        For each channel, this Ops computes the mean of the image pixels in the
        channel and then adjusts each component `x` of each pixel to
        `(x - mean) * delta + mean`.

        Tips:
            1 means pixel value no change.
            0 means all pixel equal. 
            a suitable interval is (0, 4].
        Args:
            images: a Image instance.
            delta: if int, float, a float multiplier for adjusting contrast.
                   if list, tuple, randomly picked in the interval
                   `[delta[0], delta[1])` , value is float multiplier for adjusting contrast.
        Returns:
            a Image instance.
        """
        self.image = enhance_contrast(self.image, delta)
        return self
    
    def enhance_brightness(self, delta):
        """Adjust the brightness of RGB or Grayscale images.
    
        Tips:
            delta extreme value in the interval [-1, 1], >1 to white, <-1 to black.
            a suitable interval is [-0.5, 0.5].
            0 means pixel value no change.
        Args:
            image: Tensor or array. An image.
            delta: if int, float, Amount to add to the pixel values.
                   if list, tuple, randomly picked in the interval
                   `[delta[0], delta[1])` to add to the pixel values.
        Returns:
            a Image instance.
        """
        self.image = enhance_brightness(self.image, delta)
        return self

    def enhance_sharpness(self, delta):
        """Adjust image sharpness.
        This class can be used to adjust the sharpness of an image. 
        An enhancement factor of 0.0 gives a blurred image, 
        a factor of 1.0 gives the original image, 
        and a factor of 2.0 gives a sharpened image.

        Args:
        image: a Image instance.
        delta: A floating point value controlling the enhancement. 
               delta 1.0 always returns a copy of the original image, 
               lower factors mean less sharpness, 
               and higher values more. There are no restrictions on this value.
               if list, tuple, randomly picked in the interval
                   `[delta[0], delta[1])` , value is float multiplier for adjusting color.
        Returns:
            a Image instance.
        """
        self.image = enhance_sharpness(self.image, delta)
        return self
    



from itertools import chain
from collections import defaultdict, Counter
import numpy as np
import pandas as pd

__all__ = ['CountVectorizer', 'TfidfVectorizer']

def CountVectorizer(feature, feature_scale=None, prefix='columns', dtype='int8'):
    """Convert a collection of text documents to a matrix of token counts.
    
    Args:
        feature: pd.Series or np.array or list of Lists, sample feature value.
        feature_scale: list or tuple, feature element set.
        prefix: return DataFrame columns prefix name.
        dtype: return DataFrame dtypes.
    Returns:
         CountVectorizer feature with pd.DataFrame type and feature_scale.
    """
    scale = feature_scale if feature_scale is not None else list(set(chain.from_iterable(feature)))
    class_mapping = defaultdict(int)
    class_mapping.default_factory = class_mapping.__len__
    [class_mapping[i] for i in scale]
    row = []
    col = []
    values = []
    for r, x in enumerate(feature):
        y = Counter(x)
        row.extend([r]*len(y))
        col.extend([class_mapping[i] for i in y])
        values.extend(y.values())
    t = np.zeros([max(row)+1, len(class_mapping)], dtype=dtype)
    t[row, col] = values
    if feature_scale is not None:
        t = t[:, :len(feature_scale)]
    t = pd.DataFrame(t, columns=[prefix+'_'+str(i) for i in scale])
    return t, scale

def TfidfVectorizer(count_vectorizer, norm='l2'):
    """Transform a count matrix to a normalized tf or tf-idf representation.
    
    Args:
        count_vectorizer:
        norm: 'l1', 'l2' or None, optional (default=’l2’).
              Each output row will have unit norm, either: * 'l2': Sum of squares of vector elements is 1. 
              The cosine similarity between two vectors is their dot product when l2 norm has been applied. 
              * 'l1': Sum of absolute values of vector elements is 1.
    Returns:
         sparse matrix or DataFrame, [n_samples, n_features].
    """
    t = (count_vectorizer*(np.log((count_vectorizer.shape[0]+1)/(count_vectorizer.replace({0:None}).count()+1))+1))
    if norm=='l2':
        t = (t.T/np.sqrt(np.sum(np.square(t), axis=1))).T
    elif norm=='l1':
        t = (t.T/np.sum(np.abs(t), axis=1)).T
    return t



__all__ = ['uniform', 'quantile']

def uniform(feature, bins):
    """Equal width bin, take a uniform distribution for the sample value range.
    
    Args:
        feature: pd.Series, model feature values.
        bins: int, split bins of feature.
    Returns:
        the list of split threshold of feature.
    """
    t = (feature.max()-feature.min())/bins
    m = feature.min()
    return [t*i+m for i in range(bins)]+[feature.max()]

def quantile(feature, bins):
    """Equal frequency bin, take a uniform distribution of the sample size.
    
    Args:
        feature: pd.Series, model feature values.
        bins: int, split bins of feature.
    Returns:
        the list of split threshold of feature.
    """
    t = feature.sort_values().values
    w = round(len(t)/bins)
    return [t[w*i] for i in range(bins)]+[feature.max()]



import random
import itertools
from collections import defaultdict

from joblib import Parallel, delayed

__all__ = ['RandomWalker']

class RandomWalker:
    def __init__(self, G, p=1, q=1):
        """
        :param G: networkx graph instance.
        :param p: Return parameter,controls the likelihood of immediately revisiting a node in the walk.
        :param q: In-out parameter,allows the search to differentiate between “inward” and “outward” nodes
        """
        self.G = G
        self.p = p
        self.q = q
        
        self._nodes_weight_dict = defaultdict(lambda :defaultdict(list))
        for i in self.G.nodes():
            self._nodes_weight_dict[i]['nodes'] = list(self.G.neighbors(i))
            probs = [self.G[i][j].get('weight', 1.0) for j in self._nodes_weight_dict[i]['nodes']]
            probs_sum = sum(probs)
            self._nodes_weight_dict[i]['weights'] = [i/probs_sum for i in probs]
        
#         self._edges_weight_dict = defaultdict(lambda :defaultdict(list))
#         for edge in G.edges():
#             probs = []
#             self._edges_weight_dict[edge]['nodes'] = list(self.G.neighbors(edge[1]))
#             for x in self._edges_weight_dict[edge]['nodes']:
#                 weight = self.G[edge[1]][x].get('weight', 1.0)
#                 if x == edge[0]:
#                     probs.append(weight/self.p)
#                 elif G.has_edge(x, edge[0]):
#                     probs.append(weight)
#                 else:
#                     probs.append(weight/self.q)
#             probs_sum = sum(probs)
#             self._edges_weight_dict[edge]['weights'] = [i/probs_sum for i in probs]

    def deepwalk_walks(self, walk_num, walk_length, walk_prob=False, filter_lenth=0, workers=1, verbose=0):
        return self._parallel_walks(self._deepwalk_walks, walk_num, walk_length, walk_prob, filter_lenth, workers, verbose)
        
    def deepwalk_walk(self, walk_length, start_node, walk_prob=False, filter_lenth=0):
        walk = [start_node]
        while len(walk) < walk_length:
            cur = walk[-1]
            cur_nodes = self._nodes_weight_dict[cur]['nodes']
            if len(cur_nodes) > 0:
                if walk_prob:
                    walk.append(random.choices(cur_nodes, self._nodes_weight_dict[cur]['weights'])[0])
                else:
                    walk.append(random.choice(cur_nodes))
            else:
                break
        if len(walk)<=filter_lenth:
            walk = []
        return walk
    
    def _deepwalk_walks(self, nodes, walk_num, walk_length, walk_prob, filter_lenth):
        walks = []
        for _ in range(walk_num):
            random.shuffle(nodes)
            walks += [self.deepwalk_walk(walk_length, v, walk_prob, filter_lenth) for v in nodes]
        return walks
    
    def _parallel_walks(self, method, walk_num, walk_length, walk_prob, filter_lenth, workers, verbose):
        nodes = list(self.G.nodes())
        workers_list = [walk_num//workers]*workers + ([walk_num % workers] if walk_num % workers != 0 else [])
        results = Parallel(n_jobs=workers, verbose=verbose)(
            delayed(method)(nodes, num, walk_length, walk_prob, filter_lenth) for num in workers_list)
        walks = itertools.chain(filter(lambda x: len(x) > 0, itertools.chain(*results)))
        return walks



import numpy as np
import pandas as pd

__all__ = ['normal_loss', 'mean_absolute_error', 'mean_squared_error',
           'mean_absolute_percentage_error', 'hinge', 'explained_variance_score',
           'median_absolute_error', 'r2_score', 'regression_report']

def normal_loss(y_true, y_pred, k, log=False, root=False):
    """Mean normal error regression loss.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted values, as returned by a regression.
        k: int, loss = np.sqrt(loss, 1/k).
        log: default False, whether to log the variable.
        root: default False, whether to sqrt the variable, if True, return rmse loss.
    Returns:
        regression loss values.
    """
    if log:
        loss = (np.log1p(y_true)-np.log1p(y_pred)).abs().pow(k).mean()
    else:
        loss = (y_true-y_pred).abs().pow(k).mean()
    if root:
        loss = np.sqrt(loss, 1/k)
    return loss

def mean_absolute_error(y_true, y_pred, log=False):
    """Mean absolute error regression loss.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted values, as returned by a regression.
        log: default False, whether to log the variable.
    Returns:
        regression loss values.
    """
    return normal_loss(y_true, y_pred, k=1, log=log, root=False)

def mean_squared_error(y_true, y_pred, log=False, root=False):
    """Mean squared error regression loss.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted values, as returned by a regression.
        log: default False, whether to log the variable.
        root: default False, whether to sqrt the variable, if True, return rmse loss.
    Returns:
        regression loss values.
    """
    return normal_loss(y_true, y_pred, k=2, log=log, root=root)

def mean_absolute_percentage_error(y_true, y_pred):
    """Mean absolute percentage error regression loss.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted values, as returned by a regression.
    Returns:
        regression loss values.
    """
    return ((y_true-y_pred)/y_true).abs().mean()

def hinge(y_true, y_pred, k=1):
    """hinge regression loss.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted values, as returned by a regression.
        k: int, pow() function dim.
    Returns:
        regression loss values.
    """
    return (1-y_true*y_pred).clip(lower=0).pow(k).mean()

def explained_variance_score(y_true, y_pred):
    """explained variance regression loss.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted values, as returned by a regression.
    Returns:
        regression loss values.
    """
    return 1-(y_true-y_pred).std()**2/y_true.std()**2

def median_absolute_error(y_true, y_pred):
    """Median absolute error regression loss.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted values, as returned by a regression.
    Returns:
        regression loss values.
    """
    return (y_true-y_pred).abs().median()

def r2_score(y_true, y_pred):
    """r2 regression loss.
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted values, as returned by a regression.
    Returns:
        regression loss values.
    """
    return 1-(y_true-y_pred).pow(2).sum()/(y_true-y_true.mean()).pow(2).sum()

def regression_report(y_true, y_pred, printable=False, printinfo='Regression Report'):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels.
    Returns:
        regression report.
    """
    result = {'mean_absolute_error':mean_absolute_error(y_true, y_pred),
              'mean_squared_error':mean_squared_error(y_true, y_pred),
              'mean_absolute_percentage_error':mean_absolute_percentage_error(y_true, y_pred),
              'hinge_loss':hinge(y_true, y_pred),
              'explained_variance_score':explained_variance_score(y_true, y_pred),
              'median_absolute_error':median_absolute_error(y_true, y_pred),
              'r2_score':r2_score(y_true, y_pred)
             }
    if printable:
        print("\n{}".format(printinfo))
        print("mean_absolute_error: %.4f" % result['mean_absolute_error'])
        print("mean_squared_error: %.4f" % result['mean_squared_error'])
        print("mean_absolute_percentage_error: %.4f" % result['mean_absolute_percentage_error'])
        print("hinge_loss: %.4f" % result['hinge_loss'])
        print("explained_variance_score: %.4f" % result['explained_variance_score'])
        print("median_absolute_error: %.4f" % result['median_absolute_error'])
        print("r2_score: %.4f" % result['r2_score'])
    return result



# import sys
# import time
# from collections import Counter
# from multiprocessing import cpu_count

# import numpy as np
# import lightgbm as lgb
# from sklearn.model_selection import StratifiedKFold, train_test_split


# def RandomSearch(feature, label, loss, metrics, iter_num=1000, scoring=0.5, cv=5, cv_num=3,
#                  metrics_min=True, speedy=True, speedy_param=(20000, 0.3), gpu=False):
#     """LGBClassifier model params search use RandomSearch method.
    
#     Args:
#         feature: pandas dataframe, model's feature.
#         label: pandas series, model's label.
#         loss: XGBClassifier param 'objective'.
#         metrics: model metrics function.
#         scoring: metrics error opt base line value.
#         cv: cross validation fold.
#         cv_num: minimum cross validation fold.
#         metrics_min: metrics value whether the smaller the better.
#         speedy: whether use speedy method.
#         speedy_param: if use speedy method, test_size will be set, 
#                       test_size = 1-round(min(speedy_param[0], feature.shape[0]*speedy_param[1])/feature.shape[0], 2).
#         gpu: whether use gpu.
#     Returns:
#         a best LGBClassifier model params dict.
#     Raises:
#         params error.
#     """
#     start = time.time()
#     best_params={}
#     if speedy:
#         test_size = 1-round(min(speedy_param[0], feature.shape[0]*speedy_param[1])/feature.shape[0], 2)
#     tree_method = 'gpu_hist' if gpu else 'auto'
#     n_job = 1 if gpu else int(np.ceil(cpu_count()*0.8))
#     weight_dict = Counter(label)
#     if len(weight_dict)==2:
#         weight = int(np.ceil(weight_dict[min(weight_dict)]/weight_dict[max(weight_dict)]))
#     else:
#         weight_dict = {j:i for i,j in weight_dict.items()}
#         weight = int(np.ceil(weight_dict[max(weight_dict)]/weight_dict[min(weight_dict)]))
#     for i in range(1, iter_num+1):
#         max_depth = int(np.random.choice(np.linspace(3, 8, 6)))
#         params = {'learning_rate': np.random.choice(np.linspace(0.01, 0.1, 10).round(2)),
#                   'n_estimators':,
#                   'num_leaves': int(np.random.choice(range(min(20, 2**max_depth/2), 2**max_depth))),
#                   'max_depth': max_depth,
#                   'max_bin': int(np.random.choice(range(1, 255))),
#                   'subsample_for_bin':,
#                   'min_split_gain': np.random.choice(np.linspace(0, 1, 11)).round(1),
#                   'min_child_weight': np.random.choice(np.linspace(0.0005, 0.002, 31)).round(5),
#                   'min_child_samples':,
#                   'subsample': np.random.choice(np.linspace(0.5, 1, 11)).round(2),
#                   'colsample_bytree': np.random.choice(np.linspace(0.5, 1, 11)).round(2),
#                   'reg_alpha': np.random.choice(np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2)),
#                   'reg_lambda': np.random.choice(np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2)),
#                   'random_state':27, n_jobs=n_job, 'boosting_type':'gbdt', 'objective':loss}
#         model = lgb.LGBMClassifier(**params)
#         score = []
#         if speedy:
#             for _ in range(cv_num):
#                 X_train, X_test, y_train, y_test = train_test_split(feature, label, test_size=test_size, stratify=label,
#                                                                     random_state=np.random.choice(range(100), 1)[0])
#                 model.fit(X_train, y_train)
#                 cv_pred = model.predict_proba(X_test)
#                 score.append(metrics(y_test.values, cv_pred))
#         else:
#             skf = StratifiedKFold(n_splits=cv, shuffle=True, random_state=np.random.choice(range(100), 1)[0])
#             for n, (train_index, test_index) in enumerate(skf.split(feature, label)):
#                 if n == cv_num:
#                     break
#                 model.fit(feature.loc[train_index], label[train_index])
#                 cv_pred = model.predict_proba(feature.loc[test_index])
#                 score.append(metrics(label[test_index].values, cv_pred))
#         cv_score = round(np.mean(score), 4)
#         if metrics_min:
#             if cv_score<scoring:
#                 scoring = cv_score
#                 best_params = params.copy()
#         else:
#             if cv_score>scoring:
#                 scoring = cv_score
#                 best_params = params.copy()
#         sys.stdout.write("LGBMClassifier random search percent: {}%, run time {} min, best score: {}, best param：{}\r".format(
#             round(i/iter_num*100,2), divmod((time.time()-start),60)[0], scoring, best_params))
#         sys.stdout.flush()
#     print("LGBMClassifier param finetuning with random search run time: %d min %.2f s" % divmod((time.time() - start), 60))
#     return best_params



from linora.image._image_aug import *
from linora.image._image_color import *
from linora.image._image_crop import *
from linora.image._image_io import *
from linora.image._image_noise import *
from linora.image._image_position import *
from linora.image._image_resize import *
from linora.image._image_rescale import *
from linora.image._image_util import *
from linora.image._image_filter import *



import numpy as np
import pandas as pd

__all__ = ['binary_accuracy', 'categorical_accuracy', 'recall', 'precision', 'confusion_matrix',
           'fbeta_score', 'f1_score', 'auc_roc', 'auc_pr', 'binary_crossentropy', 
           'categorical_crossentropy', 'ks', 'gini', 'psi', 'fmi', 'binary_report']

def classified_func(y_true, y_pred, prob=0.5, pos_label=1):
    t = pd.DataFrame({'prob':y_pred, 'label':y_true})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    if t.prob.nunique()!=2:
        label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
        t['label'] = t.label.replace(label_dict)
        t.loc[t.prob>=prob, 'prob'] = 1
        t.loc[t.prob<prob, 'prob'] = 0
    return t

def binary_accuracy(y_true, y_pred, prob=0.5, pos_label=1):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        the fraction of correctly classified samples (float).
    """
    t = classified_func(y_true, y_pred, prob=prob, pos_label=pos_label)
    return (t.label==t.prob).mean()

def categorical_accuracy(y_true, y_pred):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
    Returns:
        the fraction of correctly classified samples (float).
    """
    return (y_true==y_pred).mean()

def recall(y_true, y_pred, prob=0.5, pos_label=1):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        Recall of the positive class in binary classification.
    """
    t = classified_func(y_true, y_pred, prob=prob, pos_label=pos_label)
    return t.prob[t.label==pos_label].mean()

def precision(y_true, y_pred, prob=0.5, pos_label=1):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        Precision of the positive class in binary classification.
    """
    t = classified_func(y_true, y_pred, prob=prob, pos_label=pos_label)
    return t.label[t.prob==pos_label].mean()

def confusion_matrix(y_true, y_pred):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
    Returns:
        Confusion matrix.
    """
    t = pd.DataFrame({'actual':y_true, 'predict':y_pred})
    t = pd.crosstab(t.predict, t.actual)
    return t

def fbeta_score(y_true, y_pred, beta, prob=0.5, pos_label=1):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        beta : weight of precision in harmonic mean.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        Fbeta score of the positive class in binary classification.
    """
    r = recall(y_true, y_pred, prob, pos_label)
    p = precision(y_true, y_pred, prob, pos_label)
    return r*p*(1+np.power(beta, 2))/(np.power(beta, 2)*p+r)

def f1_score(y_true, y_pred, prob=0.5, pos_label=1):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        F1 score of the positive class in binary classification.
    """
    return fbeta_score(y_true, y_pred, beta=1, prob=prob, pos_label=pos_label)

def auc_roc(y_true, y_pred, pos_label=1):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        pos_label: positive label.
    Returns:
        Area Under the Receiver Operating Characteristic Curve (ROC AUC) from prediction scores.
    """
    assert y_true.nunique()==2, "`y_true` should be binary classification."
    t = pd.concat([y_true, y_pred], axis=1)
    t.columns = ['label', 'prob']
    t.insert(0, 'target', t[t.label!=pos_label].label.unique()[0])
    t = t[t.label!=pos_label].merge(t[t.label==pos_label], on='target')
    auc = (t.prob_y>t.prob_x).mean()+(t.prob_y==t.prob_x).mean()/2
    return auc

def auc_pr(y_true, y_pred, pos_label=1):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        pos_label: positive label.
    Returns:
        Area Under the Receiver Operating Characteristic Curve (PR AUC) from prediction scores.
    """
    t = pd.DataFrame({'prob':y_pred, 'label':y_true})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label'], ascending=False).reset_index(drop=True)   
    t['tp'] = t.label.cumsum()
    t['fp'] = t.index+1-t.tp
    t['recall'] = t.tp/t.label.sum()
    t['precision'] = t.tp/(t.tp+t.fp)
    auc = t.sort_values(['recall', 'precision']).drop_duplicates(['recall'], 'last').precision.mean()
    return auc

def binary_crossentropy(y_true, y_pred):
    """Computes the crossentropy metric between the labels and predictions.
    
    This is the crossentropy metric class to be used when there are only two label classes (0 and 1).
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted probability, as returned by a classifier.
    Returns:
        binary crossentropy of the positive class in binary classification.
    """
    t = np.exp(y_pred-np.max(y_pred))
    t = -(np.log(t/t.sum())*y_true).mean()
    return t

def categorical_crossentropy(y_true, y_pred, one_hot=False):
    """Computes the crossentropy metric between the labels and predictions.
    
    This is the crossentropy metric class to be used when there are multiple label classes (2 or more). 
    Here we assume that labels are given as a one_hot representation. 
    eg., When labels values are [2, 0, 1], y_true = [[0, 0, 1], [1, 0, 0], [0, 1, 0]].
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted probability, as returned by a classifier.
        one_hot: default True, Whether y_true is a one_hot variable.
    Returns:
        categorical crossentropy of the positive class in categorical classification.
    """
    assert y_pred.shape[1]==y_true.nunique(), "`y_pred` and `y_true` dim not same."
    t = np.exp(y_pred.T-np.max(y_pred, axis=1))
    if one_hot:
        t = -(np.log(t/np.sum(t, axis=0)).T*pd.get_dummies(y_true)).sum(axis=1).mean()
    else:
        t = -(np.log(t/np.sum(t, axis=0)).T*y_true).sum(axis=1).mean()
    return t

def ks(y_true, y_pred, pos_label=1):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted probability, as returned by a classifier.
        pos_label: positive label.
    Returns:
        Ks score of the positive class in binary classification.
    """
    t = pd.DataFrame({'prob':y_pred, 'label':y_true})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label'], ascending=False).reset_index(drop=True)   
    t['tp'] = t.label.cumsum()
    t['fp'] = t.index+1-t.tp
    t['tpr'] = t.tp/t.label.sum()
    t['fpr'] = t.fp/(t.label.count()-t.label.sum())
    ks = (t.tpr-t.fpr).abs().max()
    return ks

def gini(y_true, y_pred, pos_label=1):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted probability, as returned by a classifier.
        pos_label: positive label.
    Returns:
        Gini score of the positive class in binary classification.
    """
    t = pd.DataFrame({'prob':y_pred, 'label':y_true})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label'], ascending=False).reset_index(drop=True)
    gini = (t.label.cumsum().sum()/t.label.sum()-(t.label.count()+1)/2)/t.label.count()
    return gini

def psi(y_true, y_pred, threshold):
    """
    Args:
        y_true: pd.Series, a feature variable.
        y_pred: pd.Series, a feature variable.
        threshold: list, a threshold list.
    Returns:
        psi value of two variable.
    """
    actual = (y_true-y_true.min())/(y_true.max()-y_true.min())
    predict = (y_pred-y_pred.min())/(y_pred.max()-y_pred.min())
    actual = pd.cut(actual, threshold, labels=range(1, len(threshold))).value_counts(normalize=True).reset_index()
    actual.columns = ['label', 'prob1']
    predict = pd.cut(predict, threshold, labels=range(1, len(threshold))).value_counts(normalize=True).reset_index()
    predict.columns = ['label', 'prob2']
    predict = actual.merge(predict, on='label', how='outer')
    psi = ((predict.prob1-predict.prob2)*np.log((predict.prob1/(predict.prob2+0.00000001)))).sum()
    return psi

def fmi(y_true, y_pred, prob=0.5, pos_label=1):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        FMI of the positive class in binary classification.
    """
    t = classified_func(y_true, y_pred, prob=prob, pos_label=pos_label)
    t = pd.crosstab(t.label, t.prob)
    return t.iat[1, 1]/np.sqrt((t.iat[1, 1]+t.iat[1, 0])*(t.iat[1, 1]+t.iat[0, 1]))

def binary_report(y_true, y_pred, prob=0.5, pos_label=1, printable=False, printinfo='Binary Classification Report'):
    """
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted labels, as returned by a classifier.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        binary report of the positive class in binary classification.
    """
    t = classified_func(y_true, y_pred, prob=prob, pos_label=pos_label)
    cm = pd.crosstab(t.label, t.prob)
    tp = cm.iat[1, 1]
    fn = cm.iat[1, 0]
    fp = cm.iat[0, 1]
    tn = cm.iat[0, 0]
    result = {'accuracy':round((tp+tn)/(tp+tn+fp+fn), 4),
              'precision':round(tp/(tp+fp), 4),
              'recall':round(tp/(tp+fn), 4),
              'f1_score':round(2*(tp/(tp+fp))*(tp/(tp+fn))/((tp/(tp+fp))+(tp/(tp+fn))), 4),
              'auc_roc':round(auc_roc(t.label, t.prob, pos_label=pos_label), 4),
              'auc_pr':round(auc_pr(t.label, t.prob, pos_label=pos_label), 4),
              'fmi':round(tp/np.sqrt((tp+fp)*(tp+fn)), 4),
              'gini':round(gini(t.label, t.prob), 4),
              'ks':round(ks(t.label, t.prob), 4),
              'RMSE_label':round(np.sqrt(np.mean(np.square((t.label-t.prob)))), 4),
              'RMSE_prob':round(np.sqrt(np.mean(np.square((t.label-y_pred)))), 4)
             }
    if printable:
        print("\n{}".format(printinfo))
        print("Accuracy: %.4f" % result['accuracy'])
        print("Precision: %.4f" % result['precision'])
        print("Recall: %.4f" % result['recall'])
        print("F1_score: %.4f" % result['f1_score'])
        print("AUC_ROC Score: %.4f" % result['auc_roc'])
        print("AUC_PR Score: %.4f" % result['auc_pr'])
        print("FMI: %.4f" % result['fmi'])
        print("KS: %.4f" % result['ks'])
        print("Gini: %.4f" % result['gini'])
        print("RMSE_label: %.4f" % result['RMSE_label'])
        print("RMSE_prob: %.4f" % result['RMSE_prob'])
    return result



import pandas as pd

__all__  = ['mapk', 'hit_ratio', 'mean_reciprocal_rank']

def mapk(y_true, y_pred, k):
    """Mean Average Precision k
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted values, as returned by a rank.
        k: int, top k predict values.
    Returns:
        Mean Average Precision k values.
    """
    def apk(actual, predict, k):
        if len(predict)>k:
            predict = predict[:k]
        score = 0.0
        nums = 0.0
        for i,p in enumerate(predict):
            if p in actual and p not in predict[:i]:
                nums += 1.0
                score += nums / (i+1.0)
        return score / min(len(actual), k) if actual else 0.0
    return pd.DataFrame({'label1':y_true, 'label2':y_pred}).apply(lambda x:apk(x[0], x[1], k=k), axis=1).mean()

def hit_ratio(y_true, y_pred, k):
    """Hit Ratio k
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted values, as returned by a rank.
        k: int, top k predict values.
    Returns:
        Hit Ratio k values.
    """
    t = pd.DataFrame({'label1':y_true, 'label2':y_pred})
    a = t.apply(lambda x:len(set(x[0]).intersection(set(x[1]))) if len(x[1])<=k else len(set(x[0]).intersection(set(x[1][:k]))), axis=1).sum()
    b = t.label1.map(lambda x:len(set(x))).sum()
    return a/b

def mean_reciprocal_rank(y_true, y_pred, k):
    """Mean Reciprocal Rank
    
    Args:
        y_true: pd.Series, ground truth (correct) labels.
        y_pred: pd.Series, predicted values, as returned by a rank.
        k: int, top k predict values.
    Returns:
        mean reciprocal rank k values.
    """
    def mrr(actual, predict, k):
        if len(predict)>k:
            predict = predict[:k]
        try:
            rank = 1./(predict.index(actual)+1)
        except:
            rank = 0
        return rank
    return pd.DataFrame({'label1':y_true, 'label2':y_pred}).apply(lambda x: mrr(x[0], x[1], k=k), axis=1).mean()



from PIL import Image

__all__ = ['ResizeMethod', 'resize']

class resize_method:
    """For geometry operations that may map multiple input pixels to a single output pixel, 
    the Python Imaging Library provides different resampling filters.
    
    NEAREST: Pick one nearest pixel from the input image. Ignore all other input pixels.
    BOX: Each pixel of source image contributes to one pixel of the destination 
         image with identical weights. 
    BILINEAR: For resize calculate the output pixel value using linear interpolation on all 
              pixels that may contribute to the output value. 
              For other transformations linear interpolation over a 2x2 environment 
              in the input image is used.
    HAMMING: Produces a sharper image than BILINEAR, 
             doesn’t have dislocations on local level like with BOX.
    BICUBIC: For resize calculate the output pixel value using cubic interpolation on all 
             pixels that may contribute to the output value. For other transformations cubic 
             interpolation over a 4x4 environment in the input image is used.
    LANCZOS: Calculate the output pixel value using a high-quality Lanczos filter (a truncated sinc) 
             on all pixels that may contribute to the output value.
    """
    NEAREST = Image.NEAREST
    BOX = Image.BOX
    BILINEAR = Image.BILINEAR
    HAMMING = Image.HAMMING
    BICUBIC = Image.BICUBIC
    LANCZOS = Image.LANCZOS
    
ResizeMethod = resize_method()

def resize(image, size, method=ResizeMethod.BILINEAR):
    """Returns a resized copy of this image.
    
    Args:
    image: a Image instance.
    size: The requested size in pixels, as a 2-tuple: (width, height).
    method: An optional resampling filter. see la.image.ResizeMethod.
    
    Returns:
        a Image instance.
    """
    if image.mode in ['1', 'P'] and method!=ResizeMethod.NEAREST:
        method = ResizeMethod.NEAREST
    return image.resize(size, resample=method)



import random
import itertools

import pandas as pd
from linora.sample.Dataset._dataset import DataSet

__all__ = ['from_Dataframe', 'from_Array', 'from_Image']

        
class file_no(DataSet):
    def __init__(self):
        super(file_no, self).__init__()
        
    def _repeat(self):
        self.params.data_index = self.params.data_index.append([self.params.data_index for i in range(self.params.repeat_size)])
        
    def _shuffle(self):
        if isinstance(self.params.data_index, list):
            self.params.data_index = pd.Series(index=self.params.data_index, data=1).index
        if self.params.shuffle_size > 0:
            t = [self.params.data_index[self.params.shuffle_size*i:self.params.shuffle_size*(i+1)].to_list() for i in range(len(self.params.data_index)//self.params.shuffle_size+1)]
            [random.shuffle(i, random=lambda :(self.params.shuffle_seed+self.params.batch)%100/100) for i in t]
            self.params.data_index = list(itertools.chain.from_iterable(t))
        elif self.params.shuffle_size == -1:
            self.params.data_index = self.params.data_index.to_series().sample(frac=1, random_state=self.params.shuffle_seed).tolist()
            
    def _skip(self):
        self.params.data_index = self.params.data_index[self.params.skip:] if self.params.skip != -1 else []
        
    def _take(self):
        self.params.data_index = self.params.data_index[:self.params.take] if self.params.take != -1 else self.params.data_index
    
    def _shard(self):
        self.params.data_index = [self.params.data_index[i] for i in range(self.params.shard_index, len(self.params.data_index), self.params.shard_step)]

    def _batch_list_map(self, loc):
        data = list(map(self.params.map_func, *(i[loc] for i in self.params.data)))
        return [np.array(list(map(lambda x:x[i], data))) for i in range(len(data[0]))]
    
    def _batch_list(self, loc):
        return [i[loc] for i in self.params.data]
    
    def _batch_map(self, loc):
        return np.array(list(map(self.params.map_func, self.params.data[loc])))
    
    def _batch(self, loc):
        return self.params.data[loc]
    
    def __iter__(self):
        self.params.rank_list = [i[0] for i in sorted(self.params.rank.items(), key=lambda x:x[1])]
        if 'shuffle' in self.params.rank_list and 'repeat' in self.params.rank_list:
            if self.params.rank_list.index('shuffle')<self.params.rank_list.index('repeat'):
                self.params.rank_list.remove('shuffle')
                self.params.rank_list.append('shuffle')
        if  'list' in self.params.data_mode:
            if 'map' in self.params.rank_list:
                self.batch_func = self._batch_list_map
            else:
                self.batch_func = self._batch_list
        elif 'map' in self.params.rank_list:
            self.batch_func = self._batch_map
        else:
            self.batch_func = self._batch
        for i in self.params.rank_list:
            if i == 'shuffle':
                self._shuffle()
            elif i == 'repeat':
                self._repeat()
            elif i == 'skip':
                self._skip()
            elif i == 'take':
                self._take()
            elif i == 'shard':
                self._shard()
        return self
    
    def __next__(self):
        loc = self.params.data_index[self.params.batch_size*self.params.batch:self.params.batch_size*(self.params.batch+1)]
        if len(loc)==0:
            raise StopIteration
        elif len(loc)<self.params.batch_size:
            if self.params.drop_remainder:
                raise StopIteration
            else:
                self.params.batch += 1
                return self.batch_func(loc)
        self.params.batch += 1
        return self.batch_func(loc)

class from_Dataframe(file_no):
    """Represents a potentially large set of elements from dataframe."""
    def __init__(self, data):
        super(from_Dataframe, self).__init__()
        self.params.data_mode = 'Dataframe_list' if isinstance(data, tuple) else 'Dataframe'
        self.params.data_index = data[0].index if isinstance(data, tuple) else data.index
        self.params.data = [i.values for i in data] if isinstance(data, tuple) else data.values

class from_Array(file_no):
    """Represents a potentially large set of elements from array."""
    def __init__(self, data):
        super(from_Array, self).__init__()
        self.params.data_mode = 'array_list' if isinstance(data, tuple) else 'array'
        self.params.data_index = pd.Series(range(data[0].shape[0] if isinstance(data, tuple) else data.shape[0])).index
        self.params.data = data
    
class from_Image(file_no):
    """Represents a potentially large set of elements from image file list."""
    def __init__(self, data):
        super(from_Image, self).__init__()
        self.params.data_mode = 'image_list' if isinstance(data, tuple) else 'image'
        self.params.data_index = pd.Series(range(len(data[0] if isinstance(data, tuple) else data))).index
        self.params.data = [pd.Series(i).values for i in data] if isinstance(data, tuple) else pd.Series(data).values



from linora.image._image_io_aug import ImageIoAug
from linora.image._image_color_aug import ImageColorAug
from linora.image._image_noise_aug import ImageNoiseAug
from linora.image._image_resize_aug import ImageResizeAug
from linora.image._image_rescale_aug import ImageRescaleAug
from linora.image._image_position_aug import ImagePositionAug
from linora.image._image_crop_aug import ImageCropAug
from linora.image._image_filter_aug import ImageFilterAug

__all__ = ['ImageAug']

class ImageAug(ImageIoAug, ImageColorAug, ImageNoiseAug, ImageResizeAug, ImageRescaleAug, ImagePositionAug, ImageCropAug, ImageFilterAug):
    def __init__(self, image=None):
        super(ImageAug, self).__init__(image)



from collections import defaultdict
import itertools
import numpy as np

__all__ = ['HyperParametersRandom', 'HyperParametersGrid']

class HyperParametersRandom():
    def __init__(self):
        """Container for both a hyperparameter space, and current values."""
        self.params = {}
        self._space = {}
        
    def Boolean(self, name, default=None):
        """Choice between True and False.

        Arguments:
        name: Str. Name of parameter. Must be unique.
        default: Default value to return for the parameter. If unspecified, the default value will be None.
        """
        self.params[name] = np.random.choice([True, False]) if default is None else default
        if name not in self._space:
            self._space[name] = {'mode':'Boolean', 'default':default}
    
    def Float(self, name, min_value, max_value, rounds=2, default=None):
        """
        Floating point range, can be evenly divided.

        Arguments:
        name: Str. Name of parameter. Must be unique.
        min_value: Float. Lower bound of the range.
        max_value: Float. Upper bound of the range.
        rounds: Optional. Int, e.g. 2 mean round(x, 2). smallest meaningful distance between two values. 
        default: Default value to return for the parameter. If unspecified, the default value will be None.
        """
        self.params[name] = round(np.random.uniform(min_value, max_value), rounds) if default is None else default
        if name not in self._space:
            self._space[name] = {'mode':'Float', 'min_value':min_value, 'max_value':max_value, 'round':rounds, 'default':default}
    
    def Int(self, name, min_value, max_value, default=None):
        """
        Integer range.
        Note that unlinke Python's range function, max_value is included in the possible values this parameter can take on.

        Arguments:
        name: Str. Name of parameter. Must be unique.
        min_value: Int. Lower limit of range (included).
        max_value: Int. Upper limit of range (included).
        default: Default value to return for the parameter. If unspecified, the default value will be None.
        """
        self.params[name] = round(np.random.uniform(min_value, max_value)) if default is None else default
        if name not in self._space:
            self._space[name] = {'mode':'Int', 'min_value':min_value, 'max_value':max_value, 'default':default}
        
    def Choice(self, name, values, default=None):
        """
        Choice of one value among a predefined set of possible values.

        Arguments:
        name: Str. Name of parameter. Must be unique.
        values: List of possible values. Values must be int, float, str, or bool. All values must be of the same type.
        default: Default value to return for the parameter. If unspecified, the default value will be None.
        """
        self.params[name] = np.random.choice(values) if default is None else default
        if name not in self._space:
            self._space[name] = {'mode':'Choice', 'values':values, 'default':default}
    
    def update(self):
        """params update"""
        for param_name, param_config in self._space.items():
            if param_config['mode']=='Boolean':
                self.Boolean(param_name)
            if param_config['mode']=='Float':
                self.Float(param_name, param_config['min_value'], param_config['max_value'], param_config['round'])
            if param_config['mode']=='Int':
                self.Int(param_name, param_config['min_value'], param_config['max_value'])
            if param_config['mode']=='Choice':
                self.Choice(param_name, param_config['values'])
 

class HyperParametersGrid():
    def __init__(self):
        """Container for both a hyperparameter space, and current values."""
        self.params = {}
        self._space = {}
        self._rank = defaultdict(lambda:[])
        
    def Boolean(self, name, default=None, rank=0):
        """Choice between True and False.

        Arguments:
        name: Str. Name of parameter. Must be unique.
        default: Default value to return for the parameter. If unspecified, the default value will be None.
        rank: Int, default 0, Importance ordering of parameters.
        """
        self.params[name] = np.random.choice([True, False]) if default is None else default
        if name not in self._space:
            self._space[name] = {'mode':'Boolean', 'default':default, 'values':[True, False]}
            self._rank[rank].append(name)
    
    def Float(self, name, min_value, max_value, rounds=1, default=None, rank=0):
        """
        Floating point range, can be evenly divided.

        Arguments:
        name: Str. Name of parameter. Must be unique.
        min_value: Float. Lower bound of the range.
        max_value: Float. Upper bound of the range.
        rounds: Optional. Int, e.g. 2 mean round(x, 2). smallest meaningful distance between two values. 
        default: Default value to return for the parameter. If unspecified, the default value will be None.
        rank: Int, default 0, Importance ordering of parameters.
        """
        self.params[name] = round(np.random.uniform(min_value, max_value), rounds) if default is None else default
        if name not in self._space:
            self._space[name] = {'mode':'Float', 'default':default, 'values':np.linspace(min_value, max_value, round((max_value-min_value)*10**rounds)+1).round(rounds)}
            self._rank[rank].append(name)
    
    def Int(self, name, min_value, max_value, default=None, rank=0):
        """
        Integer range.
        Note that unlinke Python's range function, max_value is included in the possible values this parameter can take on.

        Arguments:
        name: Str. Name of parameter. Must be unique.
        min_value: Int. Lower limit of range (included).
        max_value: Int. Upper limit of range (included).
        default: Default value to return for the parameter. If unspecified, the default value will be None.
        rank: Int, default 0, Importance ordering of parameters.
        """
        self.params[name] = round(np.random.uniform(min_value, max_value)) if default is None else default
        if name not in self._space:
            self._space[name] = {'mode':'Int', 'default':default, 'values':np.linspace(min_value, max_value, max_value-min_value+1, dtype='int')}
            self._rank[rank].append(name)
        
    def Choice(self, name, values, default=None, rank=0):
        """
        Choice of one value among a predefined set of possible values.

        Arguments:
        name: Str. Name of parameter. Must be unique.
        values: List of possible values. Values must be int, float, str, or bool. All values must be of the same type.
        default: Default value to return for the parameter. If unspecified, the default value will be None.
        rank: Int, default 0, Importance ordering of parameters.
        """
        self.params[name] = np.random.choice(values) if default is None else default
        if name not in self._space:
            self._space[name] = {'mode':'Choice', 'default':default, 'values':values}
            self._rank[rank].append(name)
    
    def update(self, rank):
        """params update, 
        Arguments:
        rank: Int, Importance ordering of parameters.
        
        return:
        a yield.
        """
        param_list = [self._space[param_name]['values'] for param_name in self._rank[rank]]
        for value in itertools.product(*param_list):
            self.params.update({name:name_value for name, name_value in zip(self._rank[rank], value)})
            yield self.params



import math
import random
import itertools

import pandas as pd
from linora.sample.Dataset._dataset import DataSet

__all__ = ['from_Csv']

class file_yes(DataSet):
    def __init__(self):
        super(file_yes, self).__init__()
        self.params.rank_list = None
        self.params.batch_index = 0
        self.params.batch_file = 0
        self.params.sample = 0
        
    def _repeat(self):
        self.params.data = self.params.data*(self.params.repeat_size+1)
        
    def __iter__(self):
        if self.params.rank_list is None:
            self.params.rank_list = [i[0] for i in sorted(self.params.rank.items(), key=lambda x:x[1])]
        if 'repeat' in self.params.rank_list:
            self._repeat()
            self.params.rank_list.remove('repeat')
        if 'skip' not in self.params.rank_list:
            self.params.skip = None
        self.params.shuffle_size = math.ceil(max(self.params.shuffle_size, self.params.prefetch_size, 1)/self.params.batch_size)*self.params.batch_size
        self.params.df = pd.read_csv(self.params.data[self.params.batch_file], sep=self.params.sep, 
                                     iterator=True, header=self.params.header, skiprows=self.params.skip)
        self.params.values = self.params.df.get_chunk(self.params.shuffle_size).values
        self.params.data_index = list(range(self.params.values.shape[0]))
        random.shuffle(self.params.data_index, random=lambda :(self.params.shuffle_seed+self.params.batch)%100/100)
        self.params.batch_file += 1
        return self
        
    def __next__(self):
        loc = self.params.data_index[self.params.batch_size*self.params.batch_index:self.params.batch_size*(self.params.batch_index+1)]
        if len(loc)<self.params.batch_size:
            self.params.batch_index = 0
            try:
                self.params.values = self.params.df.get_chunk(self.params.shuffle_size).values
            except StopIteration:
                if self.params.batch_file==len(self.params.data):
                    raise StopIteration
                self.params.df = pd.read_csv(self.params.data[self.params.batch_file], sep=self.params.sep, header=self.params.header, iterator=True)
                self.params.values = self.params.df.get_chunk(self.params.shuffle_size).values
                self.params.batch_file += 1
            self.params.data_index = list(range(self.params.values.shape[0]))
            random.shuffle(self.params.data_index, random=lambda :(self.params.shuffle_seed+self.params.batch)%100/100)
            loc = self.params.data_index[0:self.params.batch_size]
            
        self.params.batch += 1
        self.params.batch_index += 1
        self.params.sample += len(loc)
        if self.params.take>0:
            if self.params.sample>self.params.take:
                raise StopIteration
        return self.params.values[loc]
        
class from_Csv(file_yes):
    """Represents a potentially large set of elements from csv file."""
    def __init__(self, file, sep=',', header=None):
        super(from_Csv, self).__init__()
        self.params.sep = sep
        self.params.header = header
        self.params.data = file if isinstance(file, (tuple, list)) else [file]
    



import numpy as np

__all__ = ['missing_columns', 'single_columns', 'correlation_columns',
           'cv_columns']

def missing_columns(df, missing_threshold=0.6):
    """Find missing features
    
    Parameters
    ----------
    df       : pd.DataFrame, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples and n_features is the number of features.
    missing_threshold  : float, default=0.6
        Count all features with a missing rate greater than `missing_threshold`.
    
    Returns
    -------
    t : All features with a missing rate greater than `missing_threshold`
    """
    assert 1>=missing_threshold>=0, "`missing_threshold` should be one of [0, 1]."
    t = (1-df.count()/len(df)).reset_index()
    t.columns = ['feature_name', 'missing_rate']
    t = t[t.missing_rate>=missing_threshold].reset_index(drop=True)
    return t

def single_columns(df):
    """Find single value features
    
    Parameters
    ----------
    df       : pd.DataFrame, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples and n_features is the number of features.
    
    Returns
    -------
    t : All features with single value
    """
    t = train.nunique().reset_index()
    t.columns = ['feature_name', 'single']
    t = t[t.single==1].feature_name.tolist()
    return t

def correlation_columns(df, correlation_threshold=0.9):
    """Find features whose correlation coefficient is greater than a certain threshold.
    
    Parameters
    ----------
    df       : pd.DataFrame, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples and n_features is the number of features.
    correlation_threshold : float, default=0.9
        Count all features with corr greater than `correlation_threshold`.
    
    Returns
    -------
    t : All features with corr greater than `correlation_threshold`.
    """
    t = df.corr()
    t = t.where(np.triu(np.ones(t.shape), k=1).astype(np.bool))
    return [column for column in t.columns if any(t[column].abs()>=correlation_threshold)]

def cv_columns(df, cv_threshold=0.):
    """Find features with coefficients of variation less than a certain threshold.
    
    Parameters
    ----------
    df       : pd.DataFrame, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples and n_features is the number of features.
    cv_threshold : float, default=0.
        Count all features with cv less than `cv_threshold`.
    
    Returns
    -------
    t : All features with cv less than `cv_threshold`.
    """
    t = df.replace({np.PINF:np.nan, np.NINF:np.nan})
    t = ((t-t.min())/(t.max()-t.min())*900)+100
    t = (t.std()/t.mean()).reset_index()
    t.columns = ['feature_name', 'cv_rate']
    t = t[t.cv_rate<cv_threshold].feature_name.tolist()
    return t




import numpy as np
from functools import reduce

__all__ = ['kfold', 'train_test_split']

def kfold(df, stratify=None, n_splits=3, shuffle=False, random_state=None):
    """K-Folds cross-validator
    
    Provides train/test indices to split data in train/test sets. 
    Split dataset into k consecutive folds (without shuffling by default).
    Each fold is then used once as a validation while the k - 1 remaining folds form the training set.
    
    Parameters
    ----------
    df       : pd.DataFrame, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples and n_features is the number of features.
    stratify : pd.Series, shape (n_samples,)
        The target variable for supervised learning problems.
    n_splits : int, default=3
        Number of folds. Must be at least 2.
    shuffle  : boolean, default=False, optional
        Whether to shuffle the data before splitting into batches.
    random_state : int or None, optional, default=None
        If int, random_state is the seed used by the random number generator;
        If None, the random number generator is the RandomState instance used
        Used when ``shuffle`` == True.
    
    Returns
    -------
    t : list, length=n_splits
        List each list containing train-test split of inputs.
    """
    t = df.sample(frac=1, random_state=random_state).index if shuffle else df.index
    if stratify is None:
        m = int(np.floor(len(t)/n_splits))
        fold = [t[i*m:(i+1)*m].tolist() for i in range(n_splits-1)]+[t[(n_splits-1)*m:].tolist()]
    else:
        t = stratify[t]
        fold = []
        for label in t.unique():
            a = t[t==label].index
            m = int(np.floor(len(a)/n_splits))
            fold.append([a[i*m:(i+1)*m].tolist() for i in range(n_splits-1)]+[a[(n_splits-1)*m:].tolist()])
        t = [[] for i in range(n_splits)]
        for i in fold:
            for j in range(n_splits):
                t[j] = t[j]+i[j]
        fold = t.copy()
    t = [[] for i in range(n_splits)]
    for i in range(n_splits):
        j = fold.copy()
        j.pop(i)
        j = reduce(lambda x,y:x+y, j)
        t[i] = t[i]+[j, fold[i]]
    return t

def train_test_split(df, stratify=None, test_size=0.2, shuffle=False, random_state=None):
    """Split DataFrame or matrices into random train and test subsets
    
    Parameters
    ----------
    df       : pd.DataFrame, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples and n_features is the number of features.
    stratify : pd.Series, shape (n_samples,)
        The target variable for supervised learning problems.
    test_size: float, optional (default=0.2)
        Should be between 0.0 and 1.0 and represent the proportion of the dataset to include in the test split.
    shuffle  : boolean, default=False, optional
        Whether to shuffle the data before splitting into batches.
    random_state : int or None, optional, default=None
        If int, random_state is the seed used by the random number generator;
        If None, the random number generator is the RandomState instance used
        Used when ``shuffle`` == True.
    
    Returns
    -------
    t : list, length=2
        List containing train-test split of inputs.
    """
    t = df.sample(frac=1, random_state=random_state).index if shuffle else df.index
    if stratify is None:
        t = [t[0:round(len(t)*(1-test_size))].tolist(), t[round(len(t)*(1-test_size)):].tolist()]
    else:
        t = stratify[t]
        fold = []
        for label in t.unique():
            a = t[t==label].index
            fold.append([a[0:round(len(a)*(1-test_size))].tolist(), a[round(len(a)*(1-test_size)):].tolist()])
        t = [[], []]
        for i in fold:
            for j in range(2):
                t[j] = t[j]+i[j]
        t = [sorted(t[0]), sorted(t[1])]
    return t



import sys
import time
import logging
import itertools

import numpy as np
import pandas as pd
from linora.sample_splits import kfold, train_test_split


def GridSearch(feature, label, estimator, params, cv_params, metrics, scoring=0.5, cv=5, cv_num=3,
               metrics_min=True, speedy=True, speedy_param=(20000, 0.3)):
    def product(x):
        if len(x)==1:
            return itertools.product(x[0])
        elif len(x)==2:
            return itertools.product(x[0], x[1])
        else:
            return itertools.product(x[0], x[1], x[2])
    logging.basicConfig(level=logging.ERROR)
    start = time.time()
    if speedy:
        test_size = 1-round(min(speedy_param[0], feature.shape[0]*speedy_param[1])/feature.shape[0], 2)
    for _, cv_param in cv_params.items():
        cv_param_name = [i for i in cv_param]
        cv_param_value = [cv_param[i] for i in cv_param_name]
        cv_param_iter = product(cv_param_value)
        for value in cv_param_iter:
            params.update({name:name_value for name, name_value in zip(cv_param_name, value)})
            model = xgb.XGBClassifier(**params)
            score = []
            if speedy:
                for i in range(cv_num):
                    index_list = train_test_split(feature, stratify=label, test_size=test_size,
                                              shuffle=True, random_state=np.random.choice(range(100), 1)[0])
                    model.fit(feature.loc[index_list[0]], label[index_list[0]])
                    cv_pred = pd.Series(model.predict_proba(feature.loc[index_list[1]]), index=label[index_list[1]].index)
                    score.append(metrics(label[index_list[1]], cv_pred))
            else:
                index_list = kfold(feature, stratify=label, n_splits=cv, shuffle=True, random_state=np.random.choice(range(100), 1)[0])
                for n, index in enumerate(index_list):
                    if n == cv_num:
                        break
                    model.fit(feature.loc[index[0]], label[index[0]])
                    cv_pred = pd.Series(model.predict_proba(feature.loc[index[1]]), index=label[index[1]].index)
                    score.append(metrics(label[index[1]], cv_pred))
            cv_score = round(np.mean(score), 4)
            if metrics_min:
                if cv_score<scoring:
                    scoring = cv_score
                    best_params = params.copy()
            else:
                if cv_score>scoring:
                    scoring = cv_score
                    best_params = params.copy()
        params = best_params.copy()
        sys.stdout.write("GEClassifier grid search run time {} min, best score: {}, best param：{}\r".format(
            divmod((time.time()-start),60)[0], scoring, best_params))
        sys.stdout.flush()
    print("GEClassifier param finetuning with grid search run time: %d min %.2f s" % divmod((time.time() - start), 60))
    return best_params



import numpy as np
from PIL import Image

__all__ = ['flip_up_left', 'flip_up_right', 'flip_left_right', 'flip_up_down', 'rotate', 'translate']

def flip_up_left(image, random=False):
    """Randomly flip an image (up to left).
    
    With a 1 in 2 chance, outputs the contents of `image` flipped along the
    second dimension, which is `width`.  Otherwise output the image as-is.
    
    Args:
    image: a Image instance.
    random: bool, default False.
            if True, random flip up or left image.
            if False, flip up or left image.
    Returns:
            A Image instance. of the same type and shape as `image`.
    """
    if random:
        random = np.random.choice([True, False])
    return image.transpose(Image.TRANSPOSE) if not random else image

def flip_up_right(image, random=False):
    """Randomly flip an image (up to right).
    
    With a 1 in 2 chance, outputs the contents of `image` flipped along the
    second dimension, which is `width`.  Otherwise output the image as-is.
    
    Args:
    image: a Image instance.
    random: bool, default False.
            if True, random flip up or right image.
            if False, flip up or right image.
    Returns:
            A Image instance. of the same type and shape as `image`.
    """
    if random:
        random = np.random.choice([True, False])
    return image.transpose(Image.TRANSVERSE) if not random else image

def flip_left_right(image, random=False):
    """Randomly flip an image (left to right).
    
    With a 1 in 2 chance, outputs the contents of `image` flipped along the
    second dimension, which is `width`.  Otherwise output the image as-is.
    
    Args:
    image: a Image instance.
    random: bool, default False.
            if True, random flip left or rignt image.
            if False, flip left or right image.
    Returns:
            A Image instance. of the same type and shape as `image`.
    """
    if random:
        random = np.random.choice([True, False])
    return image.transpose(Image.FLIP_LEFT_RIGHT) if not random else image

def flip_up_down(image, random=False):
    """Randomly flip an image (up to down).
    
    With a 1 in 2 chance, outputs the contents of `image` flipped along the
    second dimension, which is `width`.  Otherwise output the image as-is.
    
    Args:
    image: a Image instance.
    random: bool, default False.
            if True, random flip up or down image.
            if False, flip up or down image.
    Returns:
            A Image instance. of the same type and shape as `image`.
    """
    if random:
        random = np.random.choice([True, False])
    return image.transpose(Image.FLIP_TOP_BOTTOM) if not random else image

def rotate(image, angle, expand=True, center=None, translate=None, fillcolor=None):
    """Returns a rotated copy of this image. This method returns a copy of this image, 
    rotated the given number of degrees counter clockwise around its centre.
    
    Args:
    image: a Image instance.
    angle: In degrees counter clockwise.
           if int or float, rotation angle.
           if list or tuple, randomly picked in the interval `[angle[0], angle[1])` value.
    expand: Optional expansion flag. If true, expands the output image to make it large 
            enough to hold the entire rotated image. If false or omitted, 
            make the output image the same size as the input image. 
            Note that the expand flag assumes rotation around the center and no translation.
            if value is 'random', then the function is random.
    center: Optional center of rotation (a 2-tuple). Origin is the upper left corner. 
            Default is the center of the image.
            if value is 'random', then the function is random.
    translate: An optional post-rotate translation (a 2-tuple).
            if value is 'random', then the function is random.
    fillcolor: An optional color for area outside the rotated image.
            if value is 'random', fillcolor is one of ['green', 'red', 'white', 'black'].
            you can also pass in a list of colors.
    
    Returns:
            A Image instance. of the same type and shape as `image`.
    """
    if isinstance(angle, (list, tuple)):
        assert angle[0]<angle[1], '`angle` must be angle[0]<angle[1].'
        angle = np.random.uniform(angle[0], angle[1])
    if expand=='random':
        expand = np.random.choice([True, False])
    if center=='random':
        center = (np.random.randint(0, image.size[0]), np.random.randint(0, image.size[1]))
    if translate=='random':
        translate = (np.random.randint(0, image.size[0]*0.8), np.random.randint(0, image.size[1]*0.8))
    if fillcolor=='random':
        fillcolor = np.random.choice(['green', 'red', 'white', 'black'])
    elif isinstance(fillcolor, (list, tuple)):
        fillcolor = np.random.choice(fillcolor)
    return image.rotate(angle, expand, center, translate, fillcolor)

def translate(image, translate=None, fillcolor=None):
    """Returns a translate copy of this image. 
    
    Args:
    image: a Image instance.
    translate: An optional post-rotate translation (a 2-tuple).
            if value is 'random', then the function is random.
    fillcolor: An optional color for area outside the rotated image.
            if value is 'random', fillcolor is one of ['green', 'red', 'white', 'black'].
            you can also pass in a list of colors.
    
    Returns:
            A Image instance. of the same type and shape as `image`.
    """
    if translate=='random':
        translate = (np.random.randint(0, image.size[0]*0.8), np.random.randint(0, image.size[1]*0.8))
    if fillcolor=='random':
        fillcolor = np.random.choice(['green', 'red', 'white', 'black'])
    elif isinstance(fillcolor, (list, tuple)):
        fillcolor = np.random.choice(fillcolor)
    return rotate(image, angle=0, expand=1, center=None, translate=translate, fillcolor=fillcolor)




import numpy as np

__all__ = ['Metrics']

class Metrics:
    def __init__(self, function):
        self.function = function
        self.result = []
    
    def compute(self, y_true, y_pred):
        return self.function(y_true, y_pred)
        
    def update(self, y_true, y_pred):
        self.result += [self.function(y_true, y_pred)]
        
    def accumulate(self, latest_num=-1, method='mean'):
        if method=='mean':
            score = np.mean(self.result[-max(latest_num, 0):])
        elif method=='sum':
            score = np.sum(self.result[-max(latest_num, 0):])
        elif method=='max':
            score = np.max(self.result[-max(latest_num, 0):])
        elif method=='min':
            score = np.min(self.result[-max(latest_num, 0):])
        else:
            raise ValueError("`method` must one of `mean`, `sum`, `max`, `min`.")
        return score
    
    def reset():
        self.result = []



from functools import reduce
import itertools
from collections import defaultdict

import pandas as pd
import numpy as np

__all__ = ['categorical_encoder', 'categorical_hash', 'categorical_crossed',
           'categorical_onehot_binarizer', 'categorical_onehot_multiple']


def categorical_encoder(feature, feature_scale=None, abnormal_value=-1):
    """Encode labels with value between 0 and n_classes-1.
       
       if feature values not in feature_scale dict, return `abnormal_value`.
    
    Args:
        feature: pd.Series, sample feature.
        feature_scale: dict, label parameters dict for this estimator.
        abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
    Returns:
        return encoded labels and label parameters dict.
    """
    scale = feature_scale if feature_scale is not None else {j:i for i,j in feature.drop_duplicates().reset_index(drop=True).to_dict().items()}
    t = pd.Series([scale.get(i, abnormal_value) for i in feature], index=feature.index)
    return t, scale

def categorical_hash(feature, hash_bucket_size):
    """Hash labels with value between 0 and n_classes-1.
    
    Args:
        feature: pd.Series, sample feature.
        hash_bucket_size: int, number of categories that need hash.
    Returns:
        return hash labels.
    """
    return feature.fillna('').astype(str).map(lambda x:hash(x))%hash_bucket_size

def categorical_crossed(feature_list, hash_bucket_size):
    """Crossed categories and hash labels with value between 0 and n_classes-1.
    
    Args:
        feature_list: pd.Series list, sample feature list.
        hash_bucket_size: int, number of categories that need hash.
    Returns:
        return hash labels.
    """
    return reduce(lambda x,y:x+y, [i.fillna('').astype(str) for i in feature_list]).map(lambda x:hash(x))%hash_bucket_size

def categorical_onehot_binarizer(feature, feature_scale=None, prefix='columns', dtype='int8'):
    """Transform between iterable of iterables and a multilabel format, sample is simple categories.
    
    Args:
        feature: pd.Series, sample feature.
        feature_scale: list, feature categories list.
        prefix: String to append DataFrame column names.
        dtype: default np.uint8. Data type for new columns. Only a single dtype is allowed.
    Returns:
        Dataframe for onehot binarizer.
    """
    assert not any(feature.isnull()), "`feature' should be not contains NaN"
    scale = feature.drop_duplicates().tolist()
    if feature_scale is not None:
        t = pd.get_dummies(feature.replace({i:'temp_str' for i in set.difference(set(scale), set(feature_scale))}), prefix=prefix, dtype=dtype)
        if prefix+'_temp_str' in t.columns:
            t = t.drop([prefix+'_temp_str'], axis=1)
        for i in set.difference(set(feature_scale), set(scale)):
            if prefix+'_'+str(i) not in t.columns:
                t[prefix+'_'+str(i)] = 0
        scale = feature_scale
        t = t[[prefix+'_'+str(i) for i in feature_scale]]
    else:
        t = pd.get_dummies(feature, prefix=prefix, dtype=dtype)
        t = t[[prefix+'_'+str(i) for i in scale]]
    return t, scale

def categorical_onehot_multiple(feature, feature_scale=None, prefix='columns', dtype='int8'):
    """Transform between iterable of iterables and a multilabel format, sample is multiple categories.
    
    Args:
        feature: pd.Series, sample feature.
        feature_scale: list, feature categories list.
        prefix: String to append DataFrame column names.
        dtype: default np.uint8. Data type for new columns. Only a single dtype is allowed.
    Returns:
        Dataframe for onehot binarizer.
    """
    assert not any(feature.isnull()), "`feature' should be not contains NaN."
    scale = feature_scale if feature_scale is not None else list(set(itertools.chain.from_iterable(feature)))
    class_mapping = defaultdict(int)
    class_mapping.default_factory = class_mapping.__len__
    [class_mapping[i] for i in scale]
    col = [class_mapping[i] for i in itertools.chain.from_iterable(feature)]
    row = [i for i in itertools.chain.from_iterable(map(lambda x,y:[x]*len(y), range(len(feature)), feature))]
    t = np.zeros([max(row)+1, len(class_mapping)], dtype=dtype)
    t[row, col] = 1
    if feature_scale is not None:
        t = t[:, :len(feature_scale)]
    t = pd.DataFrame(t, columns=[prefix+'_'+str(i) for i in scale])
    return t, scale



import numpy as np

__all__ = ['normalize_global', 'normalize_channel', 'rescale']

def normalize_global(image, mean=None, std=None):
    """Normalize scales `image` to have mean and variance.
    
        This op computes `(x - mean) / std`.
        Args:
            image: An n-D Tensor where the last 3 dimensions are
                   `[height, width, channels]`.
            mean: if None, computes image mean.
                  if int or float, customize image all channels mean.
            std: if None, computes image std.
                 if int or float, customize image all channels std.
      Returns:
        The standardized image with same shape as `image`.
      Raises:
        ValueError: if the shape of 'image' is incompatible with this function.
      """
    if mean is None:
        mean = image.mean()
    elif not isinstance(mean, (int, float)):
        raise ValueError('`mean` must be int or float.')
    if std is None:
        std = image.std()
    elif not isinstance(std, (int, float)):
        raise ValueError('`std` must be int or float.')
    return (image-mean)/std

def normalize_channel(image, mean=None, std=None):
    """Normalize scales `image` to have mean and variance.
    
        This op computes `(x - mean) / std`.
        Args:
            image: An n-D Tensor where the last 3 dimensions are
                   `[height, width, channels]`.
            mean: if None, computes image mean.
                  if tuple or list, customize image each channels mean,
                  shape should 3 dims.
            std: if None, computes image std.
                 if tuple or list, customize image each channels std,
                 shape should 3 dims.
      Returns:
        The standardized image with same shape as `image`.
      Raises:
        ValueError: if the shape of 'image' is incompatible with this function.
      """
    if mean is None:
        mean = [image[:,:,i].mean() for i in range(image.shape[2])]
    elif not isinstance(mean, (list, tuple)):
        raise ValueError('`mean` must be list or tuple.')
    if std is None:
        std = [image[:,:,i].std() for i in range(image.shape[2])]
    elif not isinstance(std, (list, tuple)):
        raise ValueError('`std` must be list or tuple.')
    return (image-mean)/std

def rescale(image, scale):
    """Rescale apply to image.
    
    new pixel = image * scale
    Args:
        image: Either a 3-D float Tensor of shape [height, width, channels].
        scale: if int float, value multiply with image.
               if tuple list, randomly picked in the interval
               `[central_rate[0], central_rate[1])`, value multiply with image.
    Returns:
        The image with same shape as `image`.
    Raises:
        scale type error.
    """
    if isinstance(scale, (tuple, list)):
        scale = np.random.uniform(scale[0], scale[1])
    elif not isinstance(scale, (int, float)):
        raise ValueError('scale type should be one of int, float, tuple, list.')
    return image*scale



import numpy as np
import tensorflow as tf
from linora.image._image_io import *
from linora.image._image_color import *
from linora.image._image_noise import *
from linora.image._image_resize import *
from linora.image._image_rescale import *
from linora.image._image_position import *

__all__ = ['ImageAug']

class ImageAug():
    def __init__(self, image=None):
        """init numpy array image.
        
        Args:
            image: array, default None.
                   if image is None, you should frist use `ImageAug().read_image()`
        """
        self._image = image
    
    def read_image(self, filename, channel=0, image_format='mix'):
        """Convenience function for read image type one of `bmp`, `gif`, `jpeg`, `jpg`, and `png`.
    
        Detects whether an image is a BMP, GIF, JPEG, JPG, or PNG, and performs the
        appropriate operation to convert the input bytes `string` into a `Tensor`
        of type `dtype`.

        Note: `gif` returns a 4-D array `[num_frames, height, width, 3]`, as
        opposed to `bmp`, `jpeg`, `jpg` and `png`, which return 3-D
        arrays `[height, width, num_channels]`. Make sure to take this into account
        when constructing your graph if you are intermixing GIF files with BMP, JPEG, JPG,
        and/or PNG files.
        Args:
            filename: 0-D `string`. image absolute path.
            channels: An optional `int`. Defaults to `0`. Number of color channels for
                      the decoded image. 1 for `grayscale` and 3 for `rgb`.
            image_format: 0-D `string`. image format type one of `bmp`, `gif`, `jpeg`,
                          `jpg`, `png` and `mix`. `mix` mean contains many types image format.
        Returns:
            `Tensor` with type uint8 and shape `[height, width, num_channels]` for
            BMP, JPEG, and PNG images and shape `[num_frames, height, width, 3]` for
            GIF images.
        Raises:
            ValueError: On incorrect number of channels.
        """
        return ImageAug(read_image(filename, channel, image_format, _=True))
    
    def save_image(self, filename):
        """Writes image to the file at input filename. 
    
        Args:
            image:    A Tensor of type string. scalar. The content to be written to the output file.
            filename: A string. scalar. The name of the file to which we write the contents.
        Raises:
            ValueError: If `filename` is not in `[`jpg`, `jpeg`, `png`]`.
        """
        return save_image(self._image, filename)
    
    def RandomBrightness(self, delta, seed=None):
        """Adjust the brightness of RGB or Grayscale images.
    
        Tips:
            delta extreme value in the interval [-1, 1], >1 to white, <-1 to black.
            a suitable interval is [-0.5, 0.5].
            0 means pixel value no change.
        Args:
            delta: if int, float, Amount to add to the pixel values.
                   if list, tuple, randomly picked in the interval
                   `[delta[0], delta[1])` to add to the pixel values.
            seed: A Python integer. Used to create a random seed. See
                 `tf.set_random_seed` for behavior.
        Returns:
            A brightness-adjusted tensor of the same shape and type as `image`.
        Raises:
            ValueError: if `delta` type is error.
        """
        return ImageAug(RandomBrightness(self._image, delta, seed, _=True))
    
    def RandomContrast(self, delta, seed=None):
        """Adjust contrast of RGB or grayscale images.
    
        `images` is a tensor of at least 3 dimensions.  The last 3 dimensions are
        interpreted as `[height, width, channels]`.  The other dimensions only
        represent a collection of images, such as `[batch, height, width, channels].`

        Contrast is adjusted independently for each channel of each image.

        For each channel, this Ops computes the mean of the image pixels in the
        channel and then adjusts each component `x` of each pixel to
        `(x - mean) * delta + mean`.

        Tips:
            1 means pixel value no change.
            0 means all pixel equal. 
            a suitable interval is (0, 4].
        Args:
            delta: if int, float, a float multiplier for adjusting contrast.
                   if list, tuple, randomly picked in the interval
                   `[delta[0], delta[1])` , value is float multiplier for adjusting contrast.
            seed: A Python integer. Used to create a random seed. See
                 `tf.set_random_seed` for behavior.
        Returns:
            The contrast-adjusted image or images tensor of the same shape and type as `image`.
        Raises:
            ValueError: if `delta` type is error.
        """
        return ImageAug(RandomContrast(self._image, delta, seed, _=True))
    
    def RandomHue(self, delta, seed=None):
        """Adjust hue of an RGB image.
    
        `image` is an RGB image.  The image hue is adjusted by converting the
        image to HSV and rotating the hue channel (H) by `delta`.
        The image is then converted back to RGB.

        Tips:
            `delta` should be in the interval `[-1, 1]`, but any value is allowed.
            a suitable interval is [-0.5, 0.5].
            int value means pixel value no change.
        Args:
            delta: if float, How much to add to the hue channel.
                   if list, tuple, randomly picked in the interval
                   `[delta[0], delta[1])` , value is how much to add to the hue channel.
            seed: A Python integer. Used to create a random seed. See
                 `tf.set_random_seed` for behavior.
        Returns:
            The hue-adjusted image or images tensor of the same shape and type as `image`.
        Raises:
            ValueError: if `delta` type is error.
        """
        return ImageAug(RandomHue(self._image, delta, seed, _=True))
    
    def RandomSaturation(self, delta, seed=None):
        """Adjust saturation of an RGB image.
    
        `image` is an RGB image.  The image saturation is adjusted by converting the
        image to HSV and multiplying the saturation (S) channel by `delta` and clipping.
        The image is then converted back to RGB.

        Tips:
            if delta <= 0, image channels value are equal, image color is gray.
            a suitable interval is delta >0
        Args:
            delta: if int, float, Factor to multiply the saturation by.
                   if list, tuple, randomly picked in the interval
                   `[delta[0], delta[1])` , value is factor to multiply the saturation by.
            seed: A Python integer. Used to create a random seed. See
                 `tf.set_random_seed` for behavior.
        Returns:
            The saturation-adjusted image or images tensor of the same shape and type as `image`.
        Raises:
            ValueError: if `delta` type is error.
        """
        return ImageAug(RandomSaturation(self._image, delta, seed, _=True))
    
    def RandomGamma(self, gamma, seed=None):
        """Performs Gamma Correction on the input image.
    
        Also known as Power Law Transform. This function transforms the
        input image pixelwise according to the equation `Out = In**gamma`
        after scaling each pixel to the range 0 to 1.

        Tips:
            For gamma greater than 1, the histogram will shift towards left and
            the output image will be darker than the input image.
            For gamma less than 1, the histogram will shift towards right and
            the output image will be brighter than the input image.
            if gamma is 1, image pixel value no change.
        Args:
            gamma : if int, float, Non negative real number.
                    if list, tuple, randomly picked in the interval
                    `[delta[0], delta[1])` , value is Non negative real number.
            seed: A Python integer. Used to create a random seed. See
                  `tf.set_random_seed` for behavior.
        Returns:
            A float Tensor. Gamma corrected output image.
        Raises:
            ValueError: If gamma is negative.
        References:
            [1] http://en.wikipedia.org/wiki/Gamma_correction
        """
        return ImageAug(RandomGamma(self._image, gamma, seed, _=True))
    
    def RandomFlipLeftRight(self, random=True, seed=None):
        """Randomly flip an image horizontally (left to right).
    
        With a 1 in 2 chance, outputs the contents of `image` flipped along the
        second dimension, which is `width`.  Otherwise output the image as-is.
        Args:
            random: bool, default True.
                    if True, random flip left or rignt image.
                    if False, flip left or right image.
            seed: A Python integer. Used to create a random seed. See
                  `tf.set_random_seed` for behavior.
        Returns:
            A tensor of the same type and shape as `image`.
        Raises:
            ValueError: if the shape of `image` not supported or `random` dtype not bool.
        """
        return ImageAug(RandomFlipLeftRight(self._image, random, seed, _=True))
    
    def RandomFlipTopBottom(self, random=True, seed=None):
        """Randomly flips an image vertically (upside down).
    
        With a 1 in 2 chance, outputs the contents of `image` flipped along the first
        dimension, which is `height`.  Otherwise output the image as-is.
        Args:
            random: bool, default True.
                    if True, random flip top or bottom image.
                    if False, flip top or bottom image.
            seed: A Python integer. Used to create a random seed. See
                  `tf.set_random_seed` for behavior.
        Returns:
            A tensor of the same type and shape as `image`.
        Raises:
            ValueError: if the shape of `image` not supported or `random` dtype not bool.
        """
        return ImageAug(RandomFlipTopBottom(self._image, random, seed, _=True))
    
    def RandomTranspose(self, random=True, seed=None):
        """Transpose image(s) by swapping the height and width dimension.
    
        Args:
            random: bool, default True.
                    if True, random transpose image.
                    if False, transpose image.
            seed: A Python integer. Used to create a random seed.
                  See `tf.set_random_seed` for behavior.
        Returns:
            If `image` was 4-D, a 4-D float Tensor of shape `[batch, width, height, channels]`.
            If `image` was 3-D, a 3-D float Tensor of shape `[width, height, channels]`.
        Raises:
            ValueError: if the shape of `image` not supported or `random` dtype not bool.
        """
        return ImageAug(RandomTranspose(self._image, random, seed, _=True))
    
    def RandomRotation(self, k=[0, 1, 2, 3], seed=None):
        """Rotate image(s) counter-clockwise by 90 degrees.
    
        Tips:
            k should be int one of [1, 2, 3] or sublist in the [0, 1, 2, 3].
        Args:
            k: if k is list, random select t form k, rotation image by 90 degrees * t.
               if k is int, rotation image by 90 degrees * k.
            seed: A Python integer. Used to create a random seed.
                  See `tf.set_random_seed` for behavior.
        Returns:
        A rotated tensor of the same type and shape as `image`.
      Raises:
        ValueError: if the shape of `image` not supported or `k` dtype not int or list.
      """
        return ImageAug(RandomRotation(self._image, k, seed, _=True))
    
    def RandomCropCentralResize(self, central_rate, size, method=0, seed=None):
        """Crop the central region of the image(s) and resize specify shape.
    
        Remove the outer parts of an image but retain the central region of the image
        along each dimension. If we specify central_fraction = 0.5, this function
        returns the region marked with "X" in the below diagram.
           --------
          |        |
          |  XXXX  |
          |  XXXX  |
          |        |   where "X" is the central 50% of the image.
           --------

        This function works on either a single image (`image` is a 3-D Tensor), or a
        batch of images (`image` is a 4-D Tensor).

        Tips:
            method should be one of [0, 1, 2, 3], "0:bilinear", "1:nearest_neighbor", "2:bicubic", "3:area".
        Args:
            central_rate: if int float, should be in the interval (0, 1], fraction of size to crop.
                          if tuple list, randomly picked in the interval
                          `[central_rate[0], central_rate[1])`, value is fraction of size to crop.
            size: A 1-D int32 Tensor of 2 elements: `new_height, new_width`.
                  The new size for the images.
            method: int, default 0. resize image shape method.
                    should be one of "0:bilinear", "1:nearest_neighbor", "2:bicubic", "3:area"
            seed: A Python integer. Used to create a random seed.
                  See `tf.set_random_seed` for behavior.
        Returns:
            3-D / 4-D float Tensor, as per the input.
        Raises:
            ValueError: if central_crop_fraction is not within (0, 1].
        """
        return ImageAug(RandomCropCentralResize(self._image, central_rate, size, method, seed, _=True))
    
    def RandomCropPointResize(self, height_rate, width_rate, size, method=0, seed=None):
        """Crop the any region of the image(s) and resize specify shape.
    
        Crop region area = height_rate * width_rate *image_height * image_width

        This function works on either a single image (`image` is a 3-D Tensor), or a
        batch of images (`image` is a 4-D Tensor).

        Tips:
            method should be one of [0, 1, 2, 3], "0:bilinear", "1:nearest_neighbor", "2:bicubic", "3:area".
        Args:
            height_rate: flaot, in the interval (0, 1].
            width_rate: flaot, in the interval (0, 1].
            size: A 1-D int32 Tensor of 2 elements: `new_height, new_width`.
                  The new size for the images.
            method: int, default 0. resize image shape method.
                    should be one of "0:bilinear", "1:nearest_neighbor", "2:bicubic", "3:area"
            seed: A Python integer. Used to create a random seed.
                  See `tf.set_random_seed` for behavior.
        Returns:
            3-D / 4-D float Tensor, as per the input.
        Raises:
            ValueError: if central_crop_fraction is not within (0, 1].
        """
        return ImageAug(RandomCropPointResize(self._image, height_rate, width_rate, size, method, seed, _=True))
    
    def Normalize(self, mean=None, std=None):
        """Normalize scales `image` to have mean and variance.
    
        This op computes `(x - mean) / std`.
        Args:
            mean: if None, computes image mean.
                  if int float, customize image all channels mean.
                  if tuple list, customize image each channels mean,
                  shape should 3 dims.
            std: if None, computes image std.
                 if int float, customize image all channels std.
                 if tuple list, customize image each channels std,
                 shape should 3 dims.
        Returns:
            The standardized image with same shape as `image`.
        Raises:
            ValueError: if the shape of 'image' is incompatible with this function.
        """
        return ImageAug(Normalize(self._image, mean, std, _=True))
    
    def RandomRescale(self, scale, seed=None):
        """Rescale apply to image.
    
        new pixel = image * scale
        Args:
            scale: if int float, value multiply with image.
                   if tuple list, randomly picked in the interval
                   `[central_rate[0], central_rate[1])`, value multiply with image.
            seed: A Python integer. Used to create a random seed.
                  See `tf.set_random_seed` for behavior.
        Returns:
            3-D / 4-D float Tensor, as per the input.
        Raises:
            scale type error.
        """
        return ImageAug(RandomRescale(self._image, scale, seed, _=True))
    
    def RandomNoiseGaussian(self, scale=1, mean=0.0, std=1.0, seed=None):
        """Gaussian noise apply to image.
    
        new pixel = image + gaussian_noise * scale
        Args:
            scale: if int or float, value multiply with poisson_noise.
                   if tuple or list, randomly picked in the interval
                   `[scale[0], scale[1])`, value multiply with poisson_noise.
            mean: if int or float, value is gaussian distribution mean.
                  if tuple or list, randomly picked in the interval
                  `[mean[0], mean[1])`, value is gaussian distribution mean.
            std: if int or float, value is gaussian distribution std.
                 if tuple or list, randomly picked in the interval
                 `[std[0], std[1])`, value is gaussian distribution std.
            seed: A Python integer. Used to create a random seed.
                  See `tf.set_random_seed` for behavior.
        Returns:
            3-D / 4-D float Tensor, as per the input.
        Raises:
            scale or lam type error.
        """
        return ImageAug(RandomNoiseGaussian(self._image, scale, mean, std, seed, _=True))
    
    def RandomNoisePoisson(self, scale=1, lam=1.0, seed=None):
        """Poisson noise apply to image.
    
        new pixel = image + poisson_noise * scale
        Args:
            scale: if int or float, value multiply with poisson_noise.
                   if tuple or list, randomly picked in the interval
                   `[scale[0], scale[1])`, value multiply with poisson_noise.
            lam: if int or float, value is poisson distribution lambda.
                 if tuple or list, randomly picked in the interval
                 `[lam[0], lam[1])`, value is poisson distribution lambda.
            seed: A Python integer. Used to create a random seed.
                  See `tf.set_random_seed` for behavior.
        Returns:
            3-D / 4-D float Tensor, as per the input.
        Raises:
            scale or lam type error.
        """
        return ImageAug(RandomNoisePoisson(self._image, scale, lam, seed, _=True))
    
    def RandomNoiseMask(self, keep_prob=0.95, seed=None):
        """Mask noise apply to image.
    
        With probability `keep_prob`, outputs the input element scaled up by
        `1`, otherwise outputs `0`. 

        Tips:
            1 mean pixel have no change.
            a suitable interval is [0.9, 1].
        Args:
            image: Either a 3-D float Tensor of shape [height, width, depth], or a 4-D
                   Tensor of shape [batch_size, height, width, depth].
            keep_prob: should be in the interval (0, 1].
                       if int or float, the probability that each element is kept.
                       if tuple or list, randomly picked in the interval
                       `[keep_prob[0], keep_prob[1])`, the probability that each element is kept.
            seed: A Python integer. Used to create a random seed.
                  See `tf.set_random_seed` for behavior.
        Returns:
            3-D / 4-D float Tensor, as per the input.
        Raises:
            ValueError: If `keep_prob` is not in `(0, 1]`.
        """
        return ImageAug(RandomNoiseMask(self._image, keep_prob, seed, _=True))
    
    def RandomNoiseSaltPepper(self, keep_prob=0.95, seed=None):
        """Salt-Pepper noise apply to image.
    
        The salt-pepper noise is based on the signal-to-noise ratio of the image,
        randomly generating the pixel positions in some images all channel,
        and randomly assigning these pixels to 0 or 255.

        Tips:
            1 mean pixel have no change.
            a suitable interval is [0.9, 1].
        Args:
            image: Either a 3-D float Tensor of shape [height, width, depth], or a 4-D
                   Tensor of shape [batch_size, height, width, depth].
            keep_prob: should be in the interval (0, 1].
                       if int or float, the probability that each element is kept.
                       if tuple or list, randomly picked in the interval
                       `[keep_prob[0], keep_prob[1])`, the probability that each element is kept.
            seed: A Python integer. Used to create a random seed.
                  See `tf.set_random_seed` for behavior.
        Returns:
            3-D / 4-D float Tensor, as per the input.
        Raises:
            ValueError: If `keep_prob` is not in `(0, 1]`.
        """
        return ImageAug(RandomNoiseSaltPepper(self._image, keep_prob, seed, _=True))
    
    def RandomNoiseRainbow(self, keep_prob=0.95, seed=None):
        """Rainbowr noise apply to image.
    
        The rainbow noise is based on the signal-to-noise ratio of the image,
        randomly generating the pixel positions in some images,
        and randomly assigning these pixels to 0 or 255.

        Tips:
            1 mean pixel have no change.
            a suitable interval is [0.9, 1].
        Args:
            image: Either a 3-D float Tensor of shape [height, width, depth], or a 4-D
                   Tensor of shape [batch_size, height, width, depth].
            keep_prob: should be in the interval (0, 1].
                       if int or float, the probability that each element is kept.
                       if tuple or list, randomly picked in the interval
                       `[keep_prob[0], keep_prob[1])`, the probability that each element is kept.
            seed: A Python integer. Used to create a random seed.
                  See `tf.set_random_seed` for behavior.
        Returns:
            3-D / 4-D float Tensor, as per the input.
        Raises:
            ValueError: If `keep_prob` is not in `(0, 1]`.
        """
        return ImageAug(RandomNoiseRainbow(self._image, keep_prob, seed, _=True))
    
    def run(self):
        """return numpy array image."""
        return self._image if type(self._image)==np.ndarray else self._image.numpy()
    
    def show(self):
        """plot numpy array image."""
        return tf.keras.preprocessing.image.array_to_img((self._image if type(self._image)==np.ndarray else self._image.numpy()).astype('uint8'))



from io import open
from setuptools import setup, find_packages

def readme():
    with open('README.md', encoding='utf-8') as f:
        return f.read()

setup(name='linora',
      version='1.0.0rc3',
      install_requires=[
          'pandas>=1.0.0', 
          'pyecharts>=1.9.0',
          'pyecharts_snapshot>=0.2.0', 
          'pillow>=8.0.1',
          'joblib>=1.0.1'
      ],
      description='Simple and efficient tools for data mining and data analysis.',
      long_description=readme(),
      long_description_content_type='text/markdown',
      url='https://github.com/Hourout/linora',
      author='JinQing Lee',
      author_email='hourout@163.com',
      keywords=['hyperparameter-optimization', 'XGBoost', 'LightGBM', 'data-mining', 'data-analysis', 'machine-learning', 'image', 'text', 'data-science'],
      license='Apache License Version 2.0',
      classifiers=[
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Artificial Intelligence',
          'Topic :: Scientific/Engineering :: Visualization',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8'
      ],
      packages=find_packages(),
      zip_safe=False)
